!! ##############################################################################################
!!
!! Copyright 2012-2024 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################
!!
!! This file contains synchronous and asynchronous routines for
!! computing the QR factorization of a matrix
!!
!! ##############################################################################################

#include "qrm_common.h"


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!
!!!!  GEQRT
!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine cqrm_higeqrt(qrm_dscr, m, n, nb, ib, a, t, work, prio)

  use cqrm_dsmat_mod, alias=>cqrm_higeqrt
  use qrm_dscr_mod
  use cqrm_dsmat_task_mod

  implicit none

  type(qrm_dscr_type)   :: qrm_dscr
  type(cqrm_block_type) :: a, t
  integer               :: m, n, nb, ib
  type(cqrm_ws_type)    :: work
  integer, optional     :: prio

  integer               :: i, j, parts, kk, nn

  if(qrm_dscr%info.ne.0)      return
  if(.not.qrm_allocated(a%c)) return
  
  if(a%partitioned) then
     parts = (n-1)/nb+1
     do i=1, min(parts,(m-1)/nb+1)
        kk = min(nb,n-(i-1)*nb)
        call cqrm_higeqrt_task(qrm_dscr, m, kk, nb, ib, i, a, t, work, prio)
        do j=i+1, parts
           nn = min(nb,n-(j-1)*nb)
           call cqrm_higemqrt_task(qrm_dscr, qrm_conj_transp, m, nn, kk, nb, ib, i, j, a, t, a, work, prio)
        end do
     end do
  else
     call cqrm_higeqrt_task(qrm_dscr, m, n, nb, ib, 1, a, t, work, prio)
  end if

  return
end subroutine cqrm_higeqrt


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!
!!!!  GEMQRT
!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine cqrm_higemqrt(qrm_dscr, m, n, k, nb, ib, a, t, c, work, prio)

  use cqrm_dsmat_mod, alias=>cqrm_higemqrt
  use qrm_dscr_mod
  use cqrm_dsmat_task_mod

  implicit none

  type(qrm_dscr_type)   :: qrm_dscr
  type(cqrm_block_type) :: a, t, c
  type(cqrm_ws_type)    :: work
  integer               :: m, n, k, nb, ib
  integer, optional     :: prio

  integer               :: nn, kk, i, j, partst, partsc, nba, nbc

  if(qrm_dscr%info.ne.0)      return
  if(.not.qrm_allocated(t%c)) return
  
  if(t%partitioned) then
     partst = min((k-1)/nb+1,(m-1)/nb+1)
     nba    = nb
  else
     partst = 1
     nba    = k
  end if
  
  if(c%partitioned) then
     partsc = (n-1)/nb+1
     nbc    = nb
  else
     partsc = 1
     nbc    = n
  end if

  ! FIXME: not the cleanest. find a better way to compute the number
  ! of panels
  do i=1,  partst
     kk = min(nba, k-(i-1)*nba)
     do j=1, partsc
        nn = min(nbc, n-(j-1)*nbc)
        call cqrm_higemqrt_task(qrm_dscr, qrm_conj_transp, m, nn, kk, nba, ib, i, j, a, t, c, work, prio)
     end do
  end do

  return
end subroutine cqrm_higemqrt


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!
!!!!  TPQRT
!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine cqrm_hitpqrt(qrm_dscr, m, n, l, nb, ib, a1, a2, t, work, prio)

  use cqrm_dsmat_mod, alias=>cqrm_hitpqrt
  use qrm_dscr_mod
  use cqrm_dsmat_task_mod

  implicit none

  type(qrm_dscr_type)   :: qrm_dscr
  type(cqrm_block_type) :: a1, a2, t
  type(cqrm_ws_type)    :: work
  integer               :: m, n, l, nb, ib
  character             :: ts
  integer, optional     :: prio

  integer               :: i, j, mm, kk, nn, ll, ii, parts

  if(qrm_dscr%info.ne.0)       return
  if(.not.qrm_allocated(a2%c)) return
  
  if(a1%partitioned) then
     parts = (n-1)/nb+1
     do i=1, parts
        ii = (i-1)*nb+1
        kk = min( n-ii+1, nb )
        mm = min( m-l+ii+kk-1, m )
        if( ii.ge.l ) then
           ll = 0
        else
           ll = mm-m+l-ii+1
        end if
        call cqrm_hitpqrt_task(qrm_dscr, mm, kk, ll, nb, ib, i, a1, a2, t, work, prio)
        do j=i+1, parts
           nn = min(nb, n-(j-1)*nb)
           call cqrm_hitpmqrt_task(qrm_dscr, qrm_conj_transp,    &
                                   mm, nn, kk, ll, nb, ib, i, j, &
                                   a2, t, a1, a2,                &
                                   work, prio)
        end do
     end do
  else
     call cqrm_hitpqrt_task(qrm_dscr, m, n, l, nb, ib, 1, a1, a2, t, work, prio)
  end if

  return

end subroutine cqrm_hitpqrt


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!
!!!!  TPMQRT
!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine cqrm_hitpmqrt(qrm_dscr, m, n, k, l, nb, ib, v, t, c1, c2, work, prio)

  use cqrm_dsmat_mod, alias=>cqrm_hitpmqrt
  use qrm_dscr_mod
  use cqrm_dsmat_task_mod

  implicit none

  type(qrm_dscr_type)   :: qrm_dscr
  type(cqrm_block_type) :: v, t, c1, c2
  type(cqrm_ws_type)    :: work
  integer               :: m, n, k, l, nb, ib
  character             :: ts
  integer, optional     :: prio

  integer               :: ii, i, j, mm, nn, kk, ll, partsv, partsc, nbv, nbc

  if(qrm_dscr%info.ne.0)      return
  if(.not.qrm_allocated(v%c)) return
  
  if(v%partitioned) then
     partsv = (k-1)/nb+1
     nbv    = nb
  else
     partsv = 1
     nbv    = k
  end if

  if(c1%partitioned) then
     ! TODO: if .not. c2%partitioned raise error
     partsc = (n-1)/nb+1
     nbc    = nb
  else
     partsc = 1
     nbc    = n
  end if

  do i=1, partsv
     ii = (i-1)*nbv+1
     kk = min( k-ii+1, nbv )
     mm = min( m-l+ii+kk-1, m )
     if( ii.ge.l ) then
        ll = 0
     else
        ll = mm-m+l-ii+1
     end if
     do j=1, partsc
        nn = min(nbc, n-(j-1)*nbc)
        call cqrm_hitpmqrt_task(qrm_dscr, qrm_conj_transp,     &
                                mm, nn, kk, ll, nbv, ib, i, j, &
                                v, t, c1, c2,                  &
                                work, prio)
     end do
  end do

  return
end subroutine cqrm_hitpmqrt





! subroutine cqrm_dsmat_init_t_old(a, t, ib, bh, info)

! #if defined (have_starpu)
!   use fstarpu_mod
!   use qrm_starpu_common_mod
! #endif
!   use qrm_error_mod
!   use qrm_mem_mod
!   use qrm_parameters_mod
!   use cqrm_dsmat_mod, alias=>cqrm_dsmat_init_t
!   use qrm_dscr_mod

!   implicit none

!   type(cqrm_dsmat_type), intent(in)            :: a
!   type(cqrm_dsmat_type), intent(inout), target :: t
!   integer, intent(in)                          :: ib, bh
!   integer, optional                            :: info

!   integer                                      :: i, j, f, nbm, nbn, nbe, nbl, ibh
!   integer                                      :: m, n, mm, kk, last_mblock, last_nblock
!   logical                                      :: iseq
  
!   ! error handling
!   integer                                      :: err
!   character(len=*), parameter                  :: name='qrm_dsmat_init_t'

!   err = 0

!   if(.not.a%inited) return

!   t%seq = a%seq
!   t%pin = a%pin

!   m = a%m
!   n = a%n
  
!   nbm = cqrm_dsmat_inblock(a, m) ! number of block-rows
!   nbn = cqrm_dsmat_inblock(a, n) ! number of block-cols
  
!   nbe = min(nbm,nbn) ! number of block-eliminations
  
!   if(bh.le.0) then
!      ibh = nbm
!   else
!      ibh = bh
!   end if
  
!   if(nbe.le.0) return
  
!   ! FIXME: add error checks
!   if((ibh.ge.nbm) .or. (ibh.le.0)) then
!      ! facto has flat reduction tree (no need for T2)
!      allocate(t%blocks(nbm,nbe))
!   else
!      ! facto has hybrid reduction tree (T2 is needed)
!      allocate(t%blocks(nbm,2*nbe))
!   end if

!   do j=1, nbe
!      kk = min(a%f(j+1)-a%f(j),n-a%f(j)+1)
     
!      do i=j, nbm

!         if(.not.qrm_allocated(a%blocks(i,j)%c)) cycle

!         mm = min(a%f(i+1)-a%f(i), m-a%f(i)+1)

!         if( (mod(i-j,ibh).eq.0)) then
!            ! on this block a geqrt is done and thus we need to
!            ! allocate extra space for copying the householder
!            ! vectors. This copy allows for preventing useless
!            ! dependencies in StarPU:
!            !   |TTTT|
!            !   |1000|
!            !   |V100|
!            !   |VV10|
!            !   |VVV1|
!            call qrm_alloc(t%blocks(i,j)%c, ib+mm, kk, err, t%pin)
!            __QRM_INFO_CHECK(err, name,'qrm_alloc',9999)
  
!            if (i.ne.j ) then
!               ! the block is below the diagonal. this block is first
!               ! reduced into a triangle with a geqrf and then
!               ! annihilated with a tpqrt. Must allocate T2 for this
!               ! second operation
!               call qrm_alloc(t%blocks(i,j+nbe)%c, ib, kk, err, t%pin)
!               __QRM_INFO_CHECK(err, name,'qrm_alloc',9999)
!               t%blocks(i,j+nbe)%c = qrm_czero
!            end if
!         else
!            call qrm_alloc(t%blocks(i,j)%c, ib, kk, err, t%pin)
!            __QRM_INFO_CHECK(err, name,'qrm_alloc',9999)
!         end if

!         t%blocks(i,j)%c = qrm_czero

! #if defined (have_starpu)
!         if(.not.iseq) then
!            ! register the block to starpu
!            call fstarpu_matrix_data_register(t%blocks(i,j)%hdl, 0, &
!                 c_loc(t%blocks(i,j)%c),                            &
!                 size(t%blocks(i,j)%c,1),                           &
!                 size(t%blocks(i,j)%c,1),                           &
!                 size(t%blocks(i,j)%c,2),                           &
!                 c_sizeof(t%blocks(i,j)%c(1,1)))

!            if((mod(i-j,ibh).eq.0) .and. (i.ne.j)) then
!               ! register the block to starpu
!               call fstarpu_matrix_data_register(t%blocks(i,j+nbe)%hdl, 0, &
!                    c_loc(t%blocks(i,j+nbe)%c),                           &
!                    size( t%blocks(i,j+nbe)%c,1),                         &
!                    size( t%blocks(i,j+nbe)%c,1),                         &
!                    size( t%blocks(i,j+nbe)%c,2),                         &
!                    c_sizeof(t%blocks(i,j+nbe)%c(1,1)))
!            end if
!         end if
! #endif
!         t%blocks(i,j)%partitioned = .false.
!         if((mod(i-j,ibh).eq.0) .and. (i.ne.j)) then
!            t%blocks(i,j+nbe)%partitioned = .false.
!         end if
!      end do
!   end do

!   t%inited = .true.

! 9999 continue 
!   ! should do cleanup
!   if(present(info)) info = err
!   return
! end subroutine cqrm_dsmat_init_t_old


subroutine cqrm_dsmat_init_t(a, t, ib, bh, info)

#if defined (have_starpu)
  use fstarpu_mod
  use qrm_starpu_common_mod
#endif
  use qrm_error_mod
  use qrm_mem_mod
  use qrm_parameters_mod
  use cqrm_dsmat_mod, alias=>cqrm_dsmat_init_t
  use qrm_dscr_mod

  implicit none

  type(cqrm_dsmat_type), intent(in)            :: a
  type(cqrm_dsmat_type), intent(inout), target :: t
  integer, intent(in)                          :: ib, bh
  integer, optional                            :: info

  integer                                      :: i, j, k, l, p, nbr, nbc, nbe, ibh
  integer                                      :: m, n, mm, kk
  
  ! error handling
  integer                                      :: err
  character(len=*), parameter                  :: name='qrm_dsmat_init_t'

  err = 0

  if(.not.a%inited) return

  t%pin = a%pin
  t%seq = a%seq
  
  m = a%m
  n = a%n
  
  nbr = a%nbr
  nbc = a%nbc
  nbe = min(nbr,nbc) ! number of block-eliminations
  
  if(bh.le.0) then
     ibh = nbr
  else
     ibh = bh
  end if
  
  if(nbe.le.0) return
  
  ! FIXME: add error checks
  if((ibh.ge.nbr) .or. (ibh.le.0)) then
     ! facto has flat reduction tree (no need for T2)
     allocate(t%blocks(nbr,nbe))
  else
     ! facto has hybrid reduction tree (T2 is needed)
     allocate(t%blocks(nbr,2*nbe))
  end if

  do p=1, nbe
     kk = min(a%f(p+1)-a%f(p),n-a%f(p)+1)
     k  = p
     do l = p, nbr
        if(.not.qrm_allocated(a%blocks(l,p)%c)) exit
        
        mm = min(a%f(l+1)-a%f(l), m-a%f(l)+1)           ! this block is a killer if:
        if(  l.eq.p        .or.                        & ! it's the diagonal block or
             (l-k.ge.ibh)  .and.                       & ! it's at least bh below the last killer
             (mm.ge.kk) .and.                          & ! and it has more rows than columns
             .not. qrm_allocated(a%blocks(l,p)%stair) )& ! and does not have a staircase structure
             then
           k=l
           ! on this block a geqrt is done and thus we need to
           ! allocate extra space for copying the householder
           ! vectors. This copy allows for preventing useless
           ! dependencies in StarPU:
           !   |TTTT|
           !   |1000|
           !   |V100|
           !   |VV10|
           !   |VVV1|
           call qrm_alloc(t%blocks(l,p)%c, ib+mm, kk+ib, err, t%pin)
           __QRM_INFO_CHECK(err, name,'qrm_alloc',9999)
           t%blocks(l,p)%c = qrm_czero

           if(l.ne.p) then
              ! moreover this block will be annihilated itself and thus
              ! we must allocate a ws for this
              call qrm_alloc(t%blocks(l,p+nbe)%c, ib, kk+ib, err, t%pin)
              __QRM_INFO_CHECK(err, name,'qrm_alloc',9999)
              t%blocks(l,p+nbe)%c = qrm_czero

#if defined (have_starpu)
              if(.not.t%seq) then
                 ! register the block to starpu
                 call fstarpu_matrix_data_register(t%blocks(l,p+nbe)%hdl, 0, &
                      c_loc(t%blocks(l,p+nbe)%c),                            &
                      size( t%blocks(l,p+nbe)%c,1),                          &
                      size( t%blocks(l,p+nbe)%c,1),                          &
                      size( t%blocks(l,p+nbe)%c,2),                          &
                      c_sizeof(t%blocks(l,p+nbe)%c(1,1)))
              end if
#endif
           end if
        else
           ! this block can be annihilated using the currently active killer
           call qrm_alloc(t%blocks(l,p)%c, ib, kk+ib, err, t%pin)
           __QRM_INFO_CHECK(err, name,'qrm_alloc',9999)
        end if
        
        t%blocks(l,p)%partitioned = .false.
#if defined (have_starpu)
        if(.not.t%seq) then
           ! register the block to starpu
           call fstarpu_matrix_data_register(t%blocks(l,p)%hdl, 0, &
                c_loc(t%blocks(l,p)%c),                            &
                size( t%blocks(l,p)%c,1),                          &
                size( t%blocks(l,p)%c,1),                          &
                size( t%blocks(l,p)%c,2),                          &
                c_sizeof(t%blocks(l,p)%c(1,1)))
        end if
#endif
     end do
  end do

  t%inited = .true.

9999 continue 
  ! should do cleanup
  if(present(info)) info = err
  return
end subroutine cqrm_dsmat_init_t



subroutine cqrm_dsmat_geqr(a, t, ib, bh, work, m, n, prio, info)

  use cqrm_dsmat_mod, alias=>cqrm_dsmat_geqr
  use qrm_dscr_mod
  implicit none

  type(cqrm_dsmat_type)       :: a, t
  type(cqrm_ws_type)          :: work
  integer                     :: ib, bh
  integer, optional           :: m, n, prio
  integer, optional           :: info

  type(qrm_dscr_type)         :: qrm_dscr

  ! error handling
  integer                     :: err
  character(len=*), parameter :: name='qrm_dsmat_geqr'

  err = 0
  call qrm_dscr_init(qrm_dscr)

  call cqrm_dsmat_geqr_async(qrm_dscr, a, t, ib, bh, work, m, n, prio)

  ! wait for all the tasks in the dscriptor
  call qrm_barrier(qrm_dscr, info)

  call qrm_dscr_destroy(qrm_dscr)

9999 continue
  return
end subroutine cqrm_dsmat_geqr

subroutine cqrm_dsmat_geqr_async(qrm_dscr, a, t, ib, bh, work, m, n, prio)

  use cqrm_dsmat_mod, alias=>cqrm_dsmat_geqr_async
  use qrm_dscr_mod
  implicit none

  type(qrm_dscr_type)         :: qrm_dscr
  type(cqrm_dsmat_type)       :: a, t
  type(cqrm_ws_type)          :: work
  integer                     :: ib, bh
  integer, optional           :: m, n, prio

  integer                     :: i, j, p, s, l, nbr, nbc, nbe, first, ibh
  integer                     :: im, in, mm, nn, kk, ll, k
  
  ! error handling
  integer                     :: info
  character(len=*), parameter :: name='qrm_dsmat_geqr_async'

  if(qrm_dscr%info.ne.0) return
  info = 0

  __QRM_IFPRESENT3(im, m, a%m)
  __QRM_IFPRESENT3(in, n, a%n)
  
  if(min(im,in).eq.0) return

  nbr = cqrm_dsmat_inblock(a, im) ! number of block-rows
  nbc = cqrm_dsmat_inblock(a, in) ! number of block-cols
  nbe = min(nbr,nbc)

  if(bh.le.0) then
     ibh = nbr
  else
     ibh = bh
  end if

  if(nbe.le.0) return
  
  do p=1, nbe
     kk = min(a%f(p+1)-a%f(p),in-a%f(p)+1)
     k  = p
     ! write(*,'(" ")')
     do l = p, nbr
        if(.not.qrm_allocated(a%blocks(l,p)%c)) exit
        
        mm = min(a%f(l+1)-a%f(l), im-a%f(l)+1)           ! this block is a killer if:
        if(  l.eq.p        .or.                        & ! it's the diagonal block or
             (l-k.ge.ibh)  .and.                       & ! it's at least bh below the last killer
             (mm.ge.kk) .and.                          & ! and it has more rows than columns
             .not. qrm_allocated(a%blocks(l,p)%stair) )& ! and does not have a staircase structure
             then
           
           ! kill currently active killer using diagonal block
           if(k.ne.p) then
              ! write(*,'("tpqrt t:",i2,x,i2,"->",i2,x,i2,"  kk:",i3)')p,p,k,p,kk
              call cqrm_hitpqrt(qrm_dscr,        &
                   kk, kk, kk, kk, ib,           &
                   a%blocks(p,p),                &
                   a%blocks(k,p),                &
                   t%blocks(k,nbe+p),            &
                   work, prio)
              do j=p+1, nbc
                 nn = min(a%f(j+1)-a%f(j), in-a%f(j)+1)
                 call cqrm_hitpmqrt(qrm_dscr,     &
                      kk, nn, kk, kk, kk, ib,     &
                      a%blocks(k,p),              &
                      t%blocks(k,nbe+p),          &
                      a%blocks(p,j),              &
                      a%blocks(k,j),              &
                      work, prio)
              end do
           end if

           k = l
           ! select new killer and reduce it to triangle
           ! write(*,'("geqrt  :",i2,x,i2,"  mm:",i3," kk:",i3)')k,p,mm,kk
           call cqrm_higeqrt(qrm_dscr,              &
                mm, kk, kk, ib,                     &
                a%blocks(k,p),                      &
                t%blocks(k,p),                      &
                work, prio)
           do j=p+1, nbc
              nn = min(a%f(j+1)-a%f(j), in-a%f(j)+1)
              call cqrm_higemqrt(qrm_dscr,          &
                   mm, nn, kk, max(nn,kk), ib,      &
                   a%blocks(k,p),                   &
                   t%blocks(k,p),                   &
                   a%blocks(k,j),                   &
                   work, prio)
           end do
        else
           
           ! this block can be annihilated using the currently active killer
           ll = 0
           ! write(*,'("  tpqrt s:",i2,x,i2,"->",i2,x,i2,"  mm:",i3," kk:",i3)')k,p,l,p,mm,kk
           call cqrm_hitpqrt(qrm_dscr,           &
                mm, kk, ll, kk, ib,              &
                a%blocks(k,p),                   &
                a%blocks(l,p),                   &
                t%blocks(l,p),                   &
                work, prio)
           do j=p+1, nbc
              nn = min(a%f(j+1)-a%f(j), in-a%f(j)+1)
              call cqrm_hitpmqrt(qrm_dscr,         &
                   mm, nn, kk, ll, max(nn,kk), ib, &
                   a%blocks(l,p),                  &
                   t%blocks(l,p),                  &
                   a%blocks(k,j),                  &
                   a%blocks(l,j),                  &
                   work, prio)
           end do
           
        end if
        
     end do

     ! annihilate last currently active killer
     if(k.ne.p) then
        ! write(*,'("tpqrt t:",i2,x,i2,"->",i2,x,i2," kk:",i3)')p,p,k,p,kk
        call cqrm_hitpqrt(qrm_dscr,        &
             kk, kk, kk, kk, ib,           &
             a%blocks(p,p),                &
             a%blocks(k,p),                &
             t%blocks(k,nbe+p),            &
             work, prio)
        do j=p+1, nbc
           nn = min(a%f(j+1)-a%f(j), in-a%f(j)+1)
           call cqrm_hitpmqrt(qrm_dscr,     &
                kk, nn, kk, kk, kk, ib,     &
                a%blocks(k,p),              &
                t%blocks(k,nbe+p),          &
                a%blocks(p,j),              &
                a%blocks(k,j),              &
                work, prio)
        end do
     end if
     
  end do

9999 continue
  call qrm_error_set(qrm_dscr%info, info)
  return
end subroutine cqrm_dsmat_geqr_async





subroutine cqrm_dsmat_geqr_async_old(qrm_dscr, a, t, ib, bh, work, m, n, prio)

  use cqrm_dsmat_mod, alias=>cqrm_dsmat_geqr_async
  use qrm_dscr_mod
  implicit none

  type(qrm_dscr_type)         :: qrm_dscr
  type(cqrm_dsmat_type)       :: a, t
  type(cqrm_ws_type)          :: work
  integer                     :: ib, bh
  integer, optional           :: m, n, prio

  integer                     :: i, j, p, s, l, nbr, nbc, nbe, first, ibh
  integer                     :: im, in, mm, nn, kk, ll
  
  ! error handling
  integer                     :: info
  character(len=*), parameter :: name='qrm_dsmat_geqr_async'

  if(qrm_dscr%info.ne.0) return
  info = 0

  __QRM_IFPRESENT3(im, m, a%m)
  __QRM_IFPRESENT3(in, n, a%n)
  
  if(min(im,in).eq.0) return

  nbr = cqrm_dsmat_inblock(a, im) ! number of block-rows
  nbc = cqrm_dsmat_inblock(a, in) ! number of block-cols
  nbe = min(nbr,nbc)

  if(bh.le.0) then
     ibh = nbr
  else
     ibh = bh
  end if
  
  do p=1, nbe
     kk = min(a%f(p+1)-a%f(p),in-a%f(p)+1)
     
     ! find first sub-diagonal tile not being traversed by the staircase
     do first=p+1, nbr
        if(qrm_allocated(a%blocks(first,p)%stair)) exit
     end do

     ! Reduce the subdiagonal blocks from p to first-1 (not traversed
     ! by stair) using bh-arish tree
     do i = p, first-1, ibh
        mm = min(a%f(i+1)-a%f(i), im-a%f(i)+1)
        call cqrm_higeqrt(qrm_dscr,              &
             mm, kk, kk, ib,                     &
             a%blocks(i,p),                      &
             t%blocks(i,p),                      &
             work, prio)
        do j=p+1, nbc
           nn = min(a%f(j+1)-a%f(j), in-a%f(j)+1)
           call cqrm_higemqrt(qrm_dscr,          &
                mm, nn, kk, max(nn,kk), ib,      &
                a%blocks(i,p),                   &
                t%blocks(i,p),                   &
                a%blocks(i,j),                   &
                work, prio)
        end do
        do l=i+1, min(i+ibh-1,first-1)
           mm = min(a%f(l+1)-a%f(l), im-a%f(l)+1)
           ll = 0
           write(*,*)'tpqrt ==> ',i,p, l,p,' -- ',mm,kk,ll,kk
           call cqrm_hitpqrt(qrm_dscr,           &
                mm, kk, ll, kk, ib,              &
                a%blocks(i,p),                   &
                a%blocks(l,p),                   &
                t%blocks(l,p),                   &
                work, prio)
           do j=p+1, nbc
              nn = min(a%f(j+1)-a%f(j), in-a%f(j)+1)
              call cqrm_hitpmqrt(qrm_dscr,         &
                   mm, nn, kk, ll, max(nn,kk), ib, &
                   a%blocks(l,p),                  &
                   t%blocks(l,p),                  &
                   a%blocks(i,j),                  &
                   a%blocks(l,j),                  &
                   work, prio)
           end do
        end do
     end do
     s = ibh
     do while (s.le.first-p)
        do i = p, first-1-s, 2*s
           l = i+s
           if(l.le.first-1) then
              mm = min(a%f(l+1)-a%f(l), im-a%f(l)+1)
              ll = mm
              call cqrm_hitpqrt(qrm_dscr,        &
                   mm, kk, ll, kk, ib,           &
                   a%blocks(i,p),                &
                   a%blocks(l,p),                &
                   t%blocks(l,nbe+p),            &
                   work, prio)
              do j=p+1, nbc
                 nn = min(a%f(j+1)-a%f(j), in-a%f(j)+1)
                 call cqrm_hitpmqrt(qrm_dscr,         &
                      mm, nn, kk, ll, max(nn,kk), ib, &
                      a%blocks(l,p),                  &
                      t%blocks(l,nbe+p),              &
                      a%blocks(i,j),                  &
                      a%blocks(l,j),                  &
                      work, prio)
              end do
           end if
        end do
        s = s*2
     end do

     ! Reduce the subdiagonal blocks from first to the end (traversed
     ! by stair) using a flat tree
     do l=first, nbr
        mm = min(a%f(l+1)-a%f(l), im-a%f(l)+1)
        ll = 0
        call cqrm_hitpqrt(qrm_dscr,              &
             mm, kk, ll, kk, ib,                 &
             a%blocks(p,p),                      &
             a%blocks(l,p),                      &
             t%blocks(l,p),                      &
             work, prio)                         
        do j=p+1, nbc                             
           nn = min(a%f(j+1)-a%f(j), in-a%f(j)+1)
           call cqrm_hitpmqrt(qrm_dscr,          &
                mm, nn, kk, ll, max(nn,kk), ib,  &
                a%blocks(l,p),                   &
                t%blocks(l,p),                   &
                a%blocks(p,j),                   &
                a%blocks(l,j),                   &
                work, prio)
        end do
     end do
          
  end do

9999 continue
  call qrm_error_set(qrm_dscr%info, info)
  return
end subroutine cqrm_dsmat_geqr_async_old
