!! ##############################################################################################
!!
!! Copyright 2012-2024 CNRS, INPT
!!
!! This file is part of qr_mumps.
!!
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as
!! published by the Free Software Foundation, either version 3 of
!! the License, or (at your option) any later version.
!!
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################
!!
!! This file contains synchronous and asynchronous routines for
!! copying a matrix
!!
!! ##############################################################################################

#include "qrm_common.h"

subroutine dqrm_dsmat_extadd(a, b, i, j, m, n, l, inout, op, rowmap, colmap, prio, info)

  use dqrm_dsmat_mod, alias=>dqrm_dsmat_extadd
  use qrm_dscr_mod

  implicit none


  type(dqrm_dsmat_type)          :: a                    ! if copyout this is the source
                                                         ! matrix. If copyin it is the
                                                         ! destinantion matrix

  type(dqrm_dsmat_type)          :: b                    ! if copyout is the destination
                                                         ! matrix; if the source one.

  integer                        :: i, j, m, n, l        ! i and j are the first row
                                                         ! and column within the
                                                         ! matrix and m and n are the
                                                         ! number of rows and cols to
                                                         ! be copied. l is the number of
                                                         ! rows in the upper trapezoidal
                                                         ! part of the block (same meaning
                                                         ! as in tpqrt
  character                      :: inout, op
  integer                        :: rowmap(:,:)         ! block-row mapping from a to b
  integer, optional              :: colmap(:,:)         ! block-column mapping from a to b

  integer, optional              :: prio, info

  type(qrm_dscr_type)            :: qrm_dscr

  ! error handling
  integer                        :: err
  character(len=*), parameter    :: name='qrm_dsmat_extadd'

  err = 0

  if(.not.b%inited) then
     err = 1000
     call qrm_error_print(err, name)
     goto 9999
  end if

  call qrm_dscr_init(qrm_dscr, nocuda=.true.)

  call dqrm_dsmat_extadd_async(qrm_dscr, a, b, i=i, j=j, m=m, n=n, l=l,  &
       inout=inout, op=op,                                               &
       rowmap=rowmap, colmap=colmap, prio=prio)

  ! wait for all the tasks in the dscriptor
  call qrm_barrier(qrm_dscr, err)

  call qrm_dscr_destroy(qrm_dscr)

9999 continue
  if(present(info)) info = err
  return

end subroutine dqrm_dsmat_extadd


subroutine dqrm_dsmat_extadd_async(qrm_dscr, a, b, i, j, m, n, l, inout, op, rowmap, colmap &
#if defined(have_starpu)
     ,hdl &
#endif
     , prio)
  use dqrm_dsmat_mod, alias=>dqrm_dsmat_extadd_async
  use qrm_dscr_mod
  use qrm_mem_mod
  use iso_c_binding
  use dqrm_dsmat_task_mod

  implicit none

  type(qrm_dscr_type), target    :: qrm_dscr
  type(dqrm_dsmat_type)          :: a                    ! if copyout this is the source
                                                         ! matrix. If copyin it is the
                                                         ! destinantion matrix

  type(dqrm_dsmat_type)          :: b                    ! if copyout is the destination
                                                         ! matrix; if the source one.

  integer                        :: i, j, m, n, l        ! i and j are the first row
                                                         ! and column within the
                                                         ! matrix and m and n are the
                                                         ! number of rows and cols to
                                                         ! be copied. l is the number of
                                                         ! rows in the upper trapezoidal
                                                         ! part of the block (same meaning
                                                         ! as in tpqrt
  character                      :: inout, op
  integer                        :: rowmap(:,:)         ! block-row mapping from a to b
  integer, optional              :: colmap(:,:)         ! block-column mapping from a to b

#if defined(have_starpu)
  type(c_ptr), optional          :: hdl
#endif
  integer, optional              :: prio
  
  integer                        :: bc, br, bi, bj, bm, bn, bl
  integer                        :: first_rblock, last_rblock
  integer                        :: first_cblock, last_cblock
  
  ! error handling
  integer                        :: info
  character(len=*), parameter    :: name='qrm_dsmat_extadd_async'

  if(qrm_dscr%info.ne.0) return
  info = 0

  if(min(m,n).le.0) return

  if(.not.b%inited) then
     info = 1000
     call qrm_error_print(info, name)
     goto 9999
  end if

  ! FIXME: add check for dimensions or automatically set them

  first_rblock = dqrm_dsmat_inblock(a, i)
  first_cblock = dqrm_dsmat_inblock(a, j)
  last_rblock  = dqrm_dsmat_inblock(a, i+m-1)
  last_cblock  = dqrm_dsmat_inblock(a, j+n-1)

  do bc = first_cblock, last_cblock
     do br = first_rblock, last_rblock
        call dqrm_dsmat_block_ijmnl(a, i, j, m, n, l, br, bc, bi, bj, bm, bn, bl)
        if(min(bm,bn).gt.0) then
           call dqrm_block_extadd_task(qrm_dscr, &
                a, b, br, bc,                    &
                i=bi, j=bj, m=bm, n=bn, l=bl,    &
                inout=inout, op=op,              &
                rowmap=rowmap, colmap=colmap     &
#if defined(have_starpu)
                , hdl=hdl                        &
#endif
                , prio=prio                      &
             )
        end if
     end do
  end do

9999 continue
  call qrm_error_set(qrm_dscr%info, info)
  return
end subroutine dqrm_dsmat_extadd_async
