!! ##############################################################################################
!!
!! Copyright 2012-2024 CNRS, INPT, Inria
!!
!! This file is part of qr_mumps.
!!
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as
!! published by the Free Software Foundation, either version 3 of
!! the License, or (at your option) any later version.
!!
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################

#include "qrm_common.h"

subroutine dqrm_potrf_task(qrm_dscr, uplo, m, k, a, work, prio)

#if defined(have_starpu)
  use qrm_starpu_common_mod
  use dqrm_starpu_mod
  use qrm_pthread_mod
#endif
  use qrm_dscr_mod
  use dqrm_dsmat_mod
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  implicit none

  type(qrm_dscr_type), target :: qrm_dscr
  character, target           :: uplo
  integer, target             :: m, k
  type(dqrm_block_type)       :: a
  type(dqrm_ws_type)          :: work
  integer, optional           :: prio

  integer, target             :: iprio
  type(c_ptr)                 :: sizeof
  integer                     :: lda, i, info
  integer(kind=8)             :: ts, te
#if defined (have_starpu)
  type(c_ptr), target         :: dscrpt
  real(c_double),target       :: flops
#endif

  if(qrm_dscr%info.ne.0) return


#if defined (have_starpu)
  if(qrm_dscr%seq.eq.0) then

     iprio = qrm_starpu_get_prio(potrf_, qrm_dscr%pr, prio)

     dscrpt = c_loc(qrm_dscr)

     flops = (1.d0/3.d0)*(real(k)**3)+(real(k)**2)*real(m)+real(k)*(real(m-k)**2)

     call fstarpu_insert_task((/ dqrm_potrf_cl,                &
          FSTARPU_VALUE, c_loc(dscrpt), FSTARPU_SZ_C_PTR,      &
          FSTARPU_VALUE, c_loc(uplo),   FSTARPU_SZ_CHARACTER,  &
          FSTARPU_VALUE, c_loc(m),      FSTARPU_SZ_C_INT,      &
          FSTARPU_VALUE, c_loc(k),      FSTARPU_SZ_C_INT,      &
          FSTARPU_RW, a%hdl,                                   &
          FSTARPU_SCRATCH, work%hdl,                           &
          FSTARPU_PRIORITY, c_loc(iprio),                      &
          ! FSTARPU_FLOPS, c_loc(flops),                         &
          FSTARPU_SCHED_CTX, c_loc(qrm_dscr%ctx),              &
          C_NULL_PTR /))

  else
#endif

     ! make a call to trsm
     lda  = size(a%c,1)

     __QRM_PROF_START(ts)
     call dqrm_potrf(uplo, m, k, &
          a%c(1,1), lda, info)
     __QRM_PROF_STOP(qrm_prof_potrf_task_,ts,te)

     if(info.gt.0) then
        call qrm_atomic_cas(qrm_dscr%info, 0, 37)
        call qrm_error_print(37,'qrm_potrf')
     end if

#if defined (have_starpu)
  end if
#endif

  return

end subroutine dqrm_potrf_task


#if defined (have_starpu)
recursive subroutine dqrm_starpu_potrf_cpu_func(buffers, cl_arg) bind(C)
  use qrm_parameters_mod
  use qrm_starpu_common_mod
  use qrm_dscr_mod
  use qrm_pthread_mod
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  implicit none

  type(c_ptr), value, intent(in) :: cl_arg
  type(c_ptr), value             :: buffers

  integer(c_int), target         :: i, m, k
  integer                        :: na, lda, info
  real(r64), pointer          :: a(:,:)
  character, target              :: uplo
  type(c_ptr), target            :: dscr_c
  type(qrm_dscr_type), pointer   :: qrm_dscr
  integer(kind=8)                :: ts, te

  call fstarpu_unpack_arg(cl_arg,(/ c_loc(dscr_c), &
       c_loc(uplo),                   &
       c_loc(m),                                   &
       c_loc(k)/))

  ! until StarPU has a proper error handling method we have to check
  ! manually
  call c_f_pointer(dscr_c, qrm_dscr)
  if(qrm_dscr%info.ne.0) return

  na   = fstarpu_matrix_get_ny(buffers, 0)
  lda  = fstarpu_matrix_get_ld(buffers, 0)

  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 0), a, shape=(/lda,na/))

  __QRM_PROF_START(ts)
  call dqrm_potrf(uplo, m, k, &
       a(1,1), lda, info)
  __QRM_PROF_STOP(qrm_prof_potrf_task_,ts,te)

  if(info.gt.0) then
     call qrm_atomic_cas(qrm_dscr%info, 0, 37)
     call qrm_error_print(37,'qrm_potrf')
  end if

  return
end subroutine dqrm_starpu_potrf_cpu_func





#if defined (have_cuda) && defined (have_cusolver)
recursive subroutine dqrm_starpu_potrf_cuda_func(buffers, cl_arg) bind(C)
  use qrm_parameters_mod
  use qrm_dscr_mod
  use qrm_starpu_common_mod
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  implicit none

  type(c_ptr), value, intent(in) :: cl_arg
  type(c_ptr), value             :: buffers

  integer(c_int), target         :: m, k
  type(c_ptr), target            :: dscr_c, ap, wp
  integer                        :: ma, na, lda, nw, ldw, info, lw
  type(qrm_dscr_type), pointer   :: qrm_dscr
  integer(kind=8)                :: ts, te, tr
  character, target              :: uplo

  interface
     subroutine dqrm_cuda_potrf( &
          uplo,                  &
          m, k,                  &    
          a, lda,                &    
          w, lw) bind(C)
       use iso_c_binding
       character(c_char), value :: uplo
       type(c_ptr), value       :: a, w
       integer, value           :: m, k, lda, lw
     end subroutine dqrm_cuda_potrf
  end interface

  call fstarpu_unpack_arg(cl_arg,(/ c_loc(dscr_c), &
                                    c_loc(uplo),   &
                                    c_loc(m),      &
                                    c_loc(k)/))

  ! until StarPU has a proper error handling method we have to check
  ! manually
  call c_f_pointer(dscr_c, qrm_dscr)
  if(qrm_dscr%info.ne.0) return

  ma   = fstarpu_matrix_get_nx(buffers, 0)
  na   = fstarpu_matrix_get_ny(buffers, 0)
  lda  = fstarpu_matrix_get_ld(buffers, 0)

  nw   = fstarpu_matrix_get_ny(buffers, 1)
  ldw  = fstarpu_matrix_get_ld(buffers, 1)
  lw   = nw*ldw

  ap   = fstarpu_matrix_get_ptr(buffers, 0)
  wp   = fstarpu_matrix_get_ptr(buffers, 1)
  
  ! call system_clock(ts)
  __QRM_PROF_START(ts)
  call dqrm_cuda_potrf(         &
       uplo,                    &
       m , k ,                  &
       ap, lda,                 &
       wp, lw )
  __QRM_PROF_STOP(qrm_prof_potrf_task_,ts,te)
  ! TODO: add check for return code in info
  ! call system_clock(te,tr)
  return
end subroutine dqrm_starpu_potrf_cuda_func
#endif
#endif
