!! ##############################################################################################
!!
!! Copyright 2012-2024 CNRS, INPT
!! Copyright 2013-2015 UPS
!!
!! This file is part of qr_mumps.
!!
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as
!! published by the Free Software Foundation, either version 3 of
!! the License, or (at your option) any later version.
!!
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!> @file qrm_solve2d.F90
!! This file contains a routine that solves for R or R' against multiple vectors
!!
!!
!! ##############################################################################################


! #include "prec.h"
#include "qrm_common.h"



!> This function solves for R or R' against a single vector
!!
!! [in] qrm_spfct   the main qrm data structure after factorization.
!!
!! [in] transp      a string saying whether R or R^T will be solved
!!                  for. Only the first character is important.
!!                  
!! [in]     b       a 1d array containing the RHS vector
!!                  
!! [out]    x       a 1d array containing the solution vector
!!
subroutine sqrm_spfct_trsm1d(qrm_spfct, transp, b, x, info)

  use sqrm_spfct_mod
  use qrm_error_mod
  use sqrm_fdata_mod
  use qrm_string_mod
  use sqrm_utils_mod
  use sqrm_solve_mod, p1 => sqrm_spfct_trsm1d
  use qrm_error_mod
  implicit none

  type(sqrm_spfct_type), target  :: qrm_spfct
  real(r32), target           :: b(:)
  real(r32), optional, target :: x(:)
  character(len=*)               :: transp
  integer, optional              :: info

  real(r32), pointer          :: pntb(:,:), pntx(:,:)
  integer                        :: n, m

  n = size(b,1)
  pntb(1:n,1:1) => b(1:n)

  if(present(x)) then
     m = size(x,1)
     pntx(1:m,1:1) => x(1:m)
     call sqrm_spfct_trsm2d(qrm_spfct, transp, pntb, pntx, info)
  else
     call sqrm_spfct_trsm2d(qrm_spfct, transp, pntb, info=info)
  end if

  return

end subroutine sqrm_spfct_trsm1d




!> This function solves for R or R' against multiple vectors
!!
!! [in] qrm_spfct   the main qrm data structure after factorization.
!!
!! [in]    transp   a string saying whether R or R^T will be solved
!!                  for. Only the first character is important.
!!
!! [in]         b   a 2d array containing the RHS vectors
!!
subroutine sqrm_spfct_trsm2d(qrm_spfct, transp, b, x, info)

  use sqrm_spfct_mod
  use qrm_error_mod
  use sqrm_fdata_mod
  use sqrm_sdata_mod
  use qrm_string_mod
  use sqrm_utils_mod
  use qrm_dscr_mod
  use sqrm_solve_mod, protect => sqrm_spfct_trsm2d
  implicit none

  type(sqrm_spfct_type), target       :: qrm_spfct
  real(r32), target                :: b(:,:)
  real(r32), optional, target      :: x(:,:)
  character(len=*)                    :: transp
  integer, optional                   :: info

  type(qrm_dscr_type)                 :: qrm_dscr
  integer                             :: i, nb, nrhs, be, en, in
  integer                             :: node, inode, ret, keeph
  type(sqrm_front_type), pointer      :: front
  type(sqrm_sdata_type), allocatable  :: sdata(:)
  integer(8)                          :: ts, te1, te2, tr

  ! error management
  integer                             :: err
  character(len=*), parameter         :: name='qrm_spfct_trsm2d'

  err = 0

  ! immediately check if the facto was done. Otherwise push an error and return
  if(.not.allocated(qrm_spfct%fdata)) then
     err = 14
     call qrm_error_print(err, name)
     goto 9999
  else
     if(.not. qrm_spfct%fdata%ok) then
        err = 14
        call qrm_error_print(err, name)
        goto 9999
     end if
  end if

  call qrm_get(qrm_spfct, 'qrm_keeph', keeph)
  if(keeph .lt. 0) then
     err = 30
     call qrm_error_print(err, name)
     goto 9999
  end if

  ! blocking to deal with multiple rhs
  call qrm_get(qrm_spfct, 'qrm_rhsnb', nb)
  nrhs = size(b,2)
  if(nb.le.0) nb = nrhs

  allocate(sdata((nrhs-1)/nb+1))

  ! init the descriptor
  ! call qrm_dscr_init(qrm_dscr, nocuda=.true.)
  call qrm_dscr_init(qrm_dscr, nocuda=.false.)
  call system_clock(ts)
  do i=1, (nrhs-1)/nb+1
     if(present(x)) then
        call sqrm_sdata_init(sdata(i), qrm_spfct, &
             x(:,(i-1)*nb+1:min(i*nb,nrhs)),      &
             b(:,(i-1)*nb+1:min(i*nb,nrhs)))
     else
        call sqrm_sdata_init(sdata(i), qrm_spfct, &
             b(:,(i-1)*nb+1:min(i*nb,nrhs)),      &
             b(:,(i-1)*nb+1:min(i*nb,nrhs)))
     end if
     call sqrm_spfct_trsm_async(qrm_dscr, qrm_spfct, transp, sdata(i))
  end do
  call qrm_barrier(qrm_dscr)
  err = qrm_dscr%info
  __QRM_INFO_CHECK(err, name,'qrm_solve_async',9998)

  call qrm_dscr_destroy(qrm_dscr)

9998 continue
  do i=1, (nrhs-1)/nb+1
     call sqrm_sdata_destroy(sdata(i))
  end do
  deallocate(sdata)


9999 continue
  if(present(info)) info = err
  return

end subroutine sqrm_spfct_trsm2d


subroutine sqrm_spfct_trsm_async(qrm_dscr, qrm_spfct, transp, sdata)

  use sqrm_spfct_mod
  use sqrm_sdata_mod
  use qrm_dscr_mod
  use qrm_string_mod
  use sqrm_solve_mod, p1 => sqrm_spfct_trsm_async
  implicit none

  type(qrm_dscr_type)               :: qrm_dscr
  type(sqrm_spfct_type), intent(in) :: qrm_spfct
  character(len=*), intent(in)      :: transp
  type(sqrm_sdata_type)             :: sdata

  logical                           :: bottomup

  bottomup = (qrm_str_tolower(transp(1:1)) .eq. qrm_transp) .or. &
             (qrm_str_tolower(transp(1:1)) .eq. qrm_transp)

  if(bottomup) then
     call sqrm_spfct_trsm_bu_async(qrm_dscr, qrm_spfct, transp, sdata)
  else
     call sqrm_spfct_trsm_td_async(qrm_dscr, qrm_spfct, transp, sdata)
  end if
  
  return
  
end subroutine sqrm_spfct_trsm_async



subroutine sqrm_spfct_trsm_bu_async(qrm_dscr, qrm_spfct, transp, sdata)

  use sqrm_spfct_mod
  use qrm_error_mod
  use sqrm_fdata_mod
  use sqrm_sdata_mod
  use qrm_string_mod
  use sqrm_utils_mod
  use qrm_dscr_mod
  use sqrm_spfct_trsm_task_mod
  use sqrm_solve_mod, p1 => sqrm_spfct_trsm_bu_async
#if defined (have_starpu)
  use sqrm_starpu_mod
  use fstarpu_mod
#endif
  implicit none

  type(qrm_dscr_type)                        :: qrm_dscr
  type(sqrm_spfct_type), intent(in),  target :: qrm_spfct
  character(len=*), intent(in)               :: transp
  type(sqrm_sdata_type), target              :: sdata

  integer                                    :: i, nb, nrhs, be, en, in, br, bc
  integer                                    :: node, fnum
  type(qrm_adata_type), pointer              :: adata
  type(sqrm_fdata_type), pointer             :: fdata
  type(sqrm_front_type), pointer             :: front, child, parent
  type(sqrm_rhs_type), pointer               :: front_slv, child_rhs, parent_rhs

  ! error management
  integer                                    :: info
  character(len=*), parameter                :: name='qrm_spfct_trsm_async'

  if(qrm_dscr%info .ne. 0) return

  info = 0

  ! simplify
  adata => qrm_spfct%adata
  fdata => qrm_spfct%fdata

  nb = size(sdata%rhs,2)

  ! whether to go top-down or bottom-up
  be=1; en=adata%nnodes; in=1

  ! node is the node index (i.e., the i-th node in the give
  ! postorder) while node is its actual ID
  tree: do node=be, en, in


     ! until a proper error analysis is implemented in starpu we have
     ! to check whether an error was raised by some task in order to
     ! stop further submissions
     if(qrm_dscr%info.ne.0) goto 9999
     fnum = adata%torder(node)

     if(adata%small(fnum) .lt. 0) cycle tree
     
     front     => fdata%front_list(fnum)
     front_slv => sdata%front_slv(fnum)

     if(adata%small(fnum).gt.0) then
        call sqrm_spfct_trsm_activate_front(qrm_spfct, front, front_slv, nb, info)
        if(info.ne.0) qrm_dscr%info = info ! TODO: set atomically

        call sqrm_spfct_trsm_subtree_task(qrm_dscr, qrm_spfct, fnum, sdata, transp, 0)
__QRM_INFO_CHECK(qrm_dscr%info, name,'qrm_spfct_trsm_subtree_task',9999)

     else

        call sqrm_spfct_trsm_activate_front(qrm_spfct, front, front_slv, nb, info)
        if(info.ne.0) qrm_dscr%info = info ! TODO: set atomically
        __QRM_INFO_CHECK(qrm_dscr%info, name,'qrm_activate_front',9999)

        call sqrm_spfct_trsm_init_front_task(qrm_dscr, qrm_spfct, sdata, front%num, transp, prio=0)
        __QRM_INFO_CHECK(qrm_dscr%info, name,'qrm_init_front_task',9999)

        do br=1, front_slv%nr
           do bc=1, front_slv%nc
              if(qrm_allocated(front_slv%rhs%blocks(br,bc)%c)) then
                 call sqrm_spfct_trsm_init_block_task(qrm_dscr, front, front_slv, sdata, &
                      transp, br, bc, prio=0)
                 __QRM_INFO_CHECK(qrm_dscr%info, name,'qrm_init_task',9999)
              end if
           end do
        end do

        ! assemble front using children
        do i = adata%childptr(front%num), adata%childptr(front%num+1)-1
           child     => fdata%front_list(adata%child(i))
           child_rhs => sdata%front_slv(adata%child(i))
           ! assemble child into front
           call sqrm_spfct_trsm_assemble_front(qrm_dscr, child, child_rhs, front, front_slv, transp)
           
           do br=1, child_rhs%nr
              do bc=1, child_rhs%nc
                 call sqrm_spfct_trsm_clean_block_task(qrm_dscr, child, child_rhs, &
                      sdata, transp, br, bc, prio=0)
              end do
           end do
           
           ! once the child is assembled, it can be cleaned up
           
           call sqrm_spfct_trsm_clean_front_task(qrm_dscr, child, child_rhs, &
                sdata, transp, prio=0)
   __QRM_INFO_CHECK(qrm_dscr%info, name,'qrm_spfct_trsm_clean_task',9999)
        end do
        
        if(.not.(front%num.eq.adata%schur_node)) then
           call sqrm_dsmat_trsm_async(qrm_dscr,      &
                qrm_left, qrm_upper,                 &
                transp, qrm_no_unit,                 &
                qrm_sone,                            &
                front%f, front_slv%rhs,              &
                m=front%npiv, n=nb, k=front%n)
        end if

     end if

  end do tree

9999 continue
  call qrm_error_set(qrm_dscr%info, info)
  return

end subroutine sqrm_spfct_trsm_bu_async


subroutine sqrm_spfct_trsm_td_async(qrm_dscr, qrm_spfct, transp, sdata)

  use sqrm_spfct_mod
  use qrm_error_mod
  use sqrm_fdata_mod
  use sqrm_sdata_mod
  use qrm_string_mod
  use sqrm_utils_mod
  use qrm_dscr_mod
  use sqrm_spfct_trsm_task_mod
  use sqrm_solve_mod, p1 => sqrm_spfct_trsm_td_async
#if defined (have_starpu)
  use sqrm_starpu_mod
  use fstarpu_mod
#endif
  implicit none

  type(qrm_dscr_type)                        :: qrm_dscr
  type(sqrm_spfct_type), intent(in),  target :: qrm_spfct
  character(len=*), intent(in)               :: transp
  type(sqrm_sdata_type), target              :: sdata

  integer                                    :: i, nb, nrhs, be, en, in, br, bc
  integer                                    :: node, fnum
  type(qrm_adata_type), pointer              :: adata
  type(sqrm_fdata_type), pointer             :: fdata
  logical                                    :: bottomup
  type(sqrm_front_type), pointer             :: front, child, parent
  type(sqrm_rhs_type), pointer               :: front_slv, child_rhs, parent_rhs

  ! error management
  integer                                    :: info
  character(len=*), parameter                :: name='qrm_spfct_trsm_async'

  if(qrm_dscr%info .ne. 0) return

  info = 0

  ! simplify
  adata => qrm_spfct%adata
  fdata => qrm_spfct%fdata

  nb = size(sdata%lhs,2)

  ! whether to go top-down or bottom-up
  be=adata%nnodes; en=1; in=-1

  ! node is the node index (i.e., the i-th node in the give
  ! postorder) while node is its actual ID
  tree: do node=be, en, in
     ! until a proper error analysis is implemented in starpu we have
     ! to check whether an error was raised by some task in order to
     ! stop further submissions
     if(qrm_dscr%info.ne.0) goto 9999
     fnum = adata%torder(node)

     if(adata%small(fnum) .lt. 0) cycle tree
     
     front     => fdata%front_list(fnum)
     front_slv => sdata%front_slv(fnum)


     call sqrm_spfct_trsm_activate_front(qrm_spfct, front, front_slv, nb, info)
     if(info.ne.0) qrm_dscr%info = info ! TODO: set atomically

     call sqrm_spfct_trsm_init_front_task(qrm_dscr, qrm_spfct, sdata, front%num, transp, prio=0)
     __QRM_INFO_CHECK(qrm_dscr%info, name,'qrm_init_front_task',9999)

     do br=1, front_slv%nr
        do bc=1, front_slv%nc
           if(qrm_allocated(front_slv%rhs%blocks(br,bc)%c)) then
              call sqrm_spfct_trsm_init_block_task(qrm_dscr, front, front_slv, sdata, &
                   transp, br, bc, prio=0)
              __QRM_INFO_CHECK(qrm_dscr%info, name,'qrm_init_task',9999)
           end if
        end do
     end do

     if(adata%parent(front%num).ne.0) then
        ! assemble front using parent
        parent     => fdata%front_list(adata%parent(front%num))
        parent_rhs => sdata%front_slv(adata%parent(front%num))
        ! assemble parent into front
        call sqrm_spfct_trsm_assemble_front(qrm_dscr, front, front_slv, parent, parent_rhs, transp)
           
        if(front%num.eq.adata%child(adata%childptr(parent%num))) then
           do br=1, parent_rhs%nr
              do bc=1, parent_rhs%nc
                 call sqrm_spfct_trsm_clean_block_task(qrm_dscr, parent, parent_rhs,  &
                      sdata, transp, br, bc, prio=0)
              end do
           end do
              
           call sqrm_spfct_trsm_clean_front_task(qrm_dscr, parent, parent_rhs, &
                sdata, transp, prio=0)
        end if
     end if


     if(adata%small(fnum).gt.0) then
        call sqrm_spfct_trsm_subtree_task(qrm_dscr, qrm_spfct, fnum, &
             sdata, transp, 0)
__QRM_INFO_CHECK(qrm_dscr%info, name,'qrm_spfct_trsm_subtree_task',9999)
     else
        if(.not.(front%num.eq.adata%schur_node)) then
           call sqrm_dsmat_trsm_async(qrm_dscr,      &
                qrm_left, qrm_upper,                 &
                transp, qrm_no_unit,                 &
                qrm_sone,                            &
                front%f, front_slv%rhs,              &
                m=front%npiv, n=nb, k=front%n)
        end if
        
        if( (adata%childptr(front%num+1).eq.adata%childptr(front%num))) then
           ! In top-down, if node is a leaf then clean it up
           do br=1, front_slv%nr
              do bc=1, front_slv%nc
                 call sqrm_spfct_trsm_clean_block_task(qrm_dscr, front, front_slv, &
                      sdata, transp, br, bc, prio=0)
              end do
           end do
           
           call sqrm_spfct_trsm_clean_front_task(qrm_dscr, front, front_slv, &
                sdata, transp, prio=0)
        end if
        
     end if

  end do tree

9999 continue
  call qrm_error_set(qrm_dscr%info, info)
  return


end subroutine sqrm_spfct_trsm_td_async
