!! ##############################################################################################
!!
!! Copyright 2012-2024 CNRS, INPT
!! Copyright 2013-2015 UPS
!!
!! This file is part of qr_mumps.
!!
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as
!! published by the Free Software Foundation, either version 3 of
!! the License, or (at your option) any later version.
!!
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################

#include "qrm_common.h"

subroutine zqrm_spfct_trsm_subtree(qrm_spfct, root, sdata, transp, info)

  use zqrm_dsmat_mod
  use zqrm_spfct_mod
  use zqrm_sdata_mod
  use qrm_parameters_mod
  use qrm_string_mod
  use zqrm_solve_mod, protect=>zqrm_spfct_trsm_subtree
  implicit none

  type(zqrm_spfct_type), target  :: qrm_spfct
  type(zqrm_sdata_type), target  :: sdata
  integer                        :: root
  character(len=*), intent(in)   :: transp
  integer, optional              :: info

  logical                        :: bottomup
  type(qrm_adata_type), pointer  :: adata
  type(zqrm_fdata_type), pointer :: fdata
  type(qrm_dscr_type)            :: qrm_seq_dscr
  integer                        :: node, f, nb, bc, br, c, endnode, inc
  type(zqrm_front_type), pointer :: front, child, parent
  type(zqrm_rhs_type), pointer   :: front_slv, child_rhs, parent_rhs

  ! error management
  integer                        :: err
  character(len=*), parameter    :: name='qrm_spfct_trsm_subtree'

  err = 0

  ! simplify
  adata => qrm_spfct%adata
  fdata => qrm_spfct%fdata

  nb = size(sdata%lhs,2)

  ! whether to go top-down or bottom-up
  bottomup = (qrm_str_tolower(transp(1:1)) .eq. qrm_conj_transp) .or. &
             (qrm_str_tolower(transp(1:1)) .eq. qrm_transp)

  if(bottomup) then
     ! pointer to the first leaf of the sequential subtree
     node    = adata%small(root)
     endnode = root
     inc     = 1
  else
     node    = root
     endnode = adata%small(root)
     inc     = -1
  end if

  
  call qrm_dscr_init(qrm_seq_dscr, seq=.true.)

  subtree: do
     f         =  adata%torder(node)
     front     => fdata%front_list(f)
     front_slv => sdata%front_slv(f)

     if(front%num.ne.root) then
        call zqrm_spfct_trsm_activate_front(qrm_spfct, front, front_slv, nb, err)
        __QRM_INFO_CHECK(err, name,'qrm_spfct_trsm_activate_front',9998)
     end if
        
     if(bottomup .or. front%num.ne.root) then
        call zqrm_spfct_trsm_init_front(qrm_spfct, front, front_slv, err)
        
        do br=1, front_slv%nr
           do bc=1, front_slv%nc
              if(qrm_allocated(front_slv%rhs%blocks(br,bc)%c)) then
                 call zqrm_spfct_trsm_init_block(front, front_slv, &
                      transp, br, bc, sdata%rhs, err)
              end if
           end do
        end do
     end if

     if(bottomup) then
        do c = adata%childptr(f), adata%childptr(f+1)-1
           child     => fdata%front_list(adata%child(c))
           child_rhs => sdata%front_slv(child%num)
           ! the double condition below should be redundant but you never know...
           call zqrm_spfct_trsm_assemble_front(qrm_seq_dscr, child, child_rhs, front, front_slv, transp)
           do br=1, child_rhs%nr
              do bc=1, child_rhs%nc
                 call zqrm_spfct_trsm_clean_block(child, child_rhs, &
                      transp, br, bc, sdata%lhs, err)
              end do
           end do

           call zqrm_spfct_trsm_clean_front(child, child_rhs, sdata%lhs, transp, err)
           ! if(child%m.gt.child%n) sdata%p(child%rows(child%n+1:child%m),:) = qrm_zzero
           ! call zqrm_dsmat_destroy(child_rhs%rhs, seq=.true.)

        end do
     else if(.not.bottomup .and. front%num.ne.root) then
        parent     => fdata%front_list(adata%parent(front%num))
        parent_rhs => sdata%front_slv(adata%parent(front%num))
        ! the double condition below should be redundant but you never know...
        call zqrm_spfct_trsm_assemble_front(qrm_seq_dscr, front, front_slv, parent, parent_rhs, transp)

        if(front%num.eq.adata%child(adata%childptr(parent%num))) then
           do br=1, parent_rhs%nr
              do bc=1, parent_rhs%nc
                 call zqrm_spfct_trsm_clean_block(parent, parent_rhs,  &
                      transp, br, bc, sdata%lhs, err)
              end do
           end do

           call zqrm_spfct_trsm_clean_front(parent, parent_rhs, sdata%lhs, transp, err)
        end if
     end if


     ! write(*,'("+++++++++++++++++++++++++++++++++++++++++++++++++++")')
     ! call zqrm_dsmat_print(front%f, 'f15.10')
     ! write(*,'("+++++++++++++++++++++++++++++++++++++++++++++++++++")')
     ! call zqrm_dsmat_print(front_slv%rhs, 'f15.10')
     ! write(*,'(" ")')
     ! write(*,'("+++++++++++++++++++++++++++++++++++++++++++++++++++")')
     ! write(*,*)'---> ',front%num,front%m,front%n,front%npiv
     if(.not.(front%num.eq.adata%schur_node)) then
        call zqrm_dsmat_trsm_async(qrm_seq_dscr,  &
             qrm_left, qrm_upper,                 &
             transp, qrm_no_unit,                 &
             qrm_zone,                            &
             front%f, front_slv%rhs,              &
             m=front%npiv, n=nb, k=front%n)
     end if
        
     ! call zqrm_dsmat_print(front_slv%rhs, 'f15.10')
     ! write(*,'("+++++++++++++++++++++++++++++++++++++++++++++++++++")')


     if(  .not.bottomup .and. &
          (adata%childptr(front%num+1).eq.adata%childptr(front%num))) then
        ! In top-down, if node is a leaft then clean it up
        do br=1, front_slv%nr
           do bc=1, front_slv%nc
              call zqrm_spfct_trsm_clean_block(front, front_slv, &
                   transp, br, bc, sdata%lhs, err)
           end do
        end do

        call zqrm_spfct_trsm_clean_front(front, front_slv, sdata%lhs, transp, err)
     end if

     if(front%num.eq.endnode) exit subtree
     node = node+inc

  end do subtree

  err = qrm_seq_dscr%info

9998 continue ! error management
  if(present(info)) info = err
  return
end subroutine zqrm_spfct_trsm_subtree
