!! ##############################################################################################
!!
!! Copyright 2012-2024 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!> @file qrm_spmat_ata.F90
!! This file contains the routine that computes  A^T * A
!!
!!
!! ##############################################################################################


#include "qrm_common.h"

!> @brief This subroutine computes the A^T * A  product for an input matrix in csc format
!! 
!! @param[in] a_csc  the A matrix (in CSC format)
!! 
!! @param[out] ata_csr  the A'*A matrix in CSR format
!! 
subroutine cqrm_spmat_ata(a_csc, ata_csr, info)
  use cqrm_spmat_mod, alias=>cqrm_spmat_ata
  use qrm_error_mod
  use qrm_mem_mod

  implicit none

  type(cqrm_spmat_type), intent(in)  :: a_csc
  type(cqrm_spmat_type), intent(out) :: ata_csr
  integer, optional                  :: info
  
  type(cqrm_spmat_type)              :: a_csr
  integer                            :: i, j, row1, col1, row2, col2
  integer, allocatable               :: mark(:), idx(:)
  ! complex(r32), allocatable          :: x
  
  ! error management
  integer                            :: err, err2
  character(len=*), parameter        :: name='qrm_spmat_ata'
  
  err = 0; err2 = 0

  call cqrm_spmat_convert(a_csc, a_csr, 'csr', info=err)
  __QRM_INFO_CHECK(err, name, 'qrm_spmat_convert', 9999)

  call qrm_alloc(ata_csr%iptr,a_csc%n+2, err)      ! one more just to make code cleaner
  __QRM_INFO_CHECK(err, name, 'qrm_alloc', 9999)
  ata_csr%iptr = 0
  ata_csr%iptr(1:2) = 1

  call qrm_alloc(mark, a_csc%n, err)
  __QRM_INFO_CHECK(err, name, 'qrm_alloc', 9999)

  call qrm_alloc(idx, a_csc%n, err)
  __QRM_INFO_CHECK(err, name, 'qrm_alloc', 9999)

  mark = 0

  do col1=1, a_csc%n
     ! loop over all the columns of A
     do i=a_csc%jptr(col1), a_csc%jptr(col1+1)-1
        ! for each nnz in the column, we go through the corresponding
        ! row and count all the i,j pairs
        row1 = a_csc%irn(i)

        do j=a_csr%iptr(row1), a_csr%iptr(row1+1)-1
           col2 = a_csr%jcn(j)
           ! now, the element (col1,col2) is present in A'*A. Check if
           ! it was already found, otherwise add it.
           ! skip the diagonal
           if(col1 .gt. col2) cycle
           if(mark(col2) .lt. col1) then
              mark(col2) = col1
              ata_csr%iptr(col1+2) = ata_csr%iptr(col1+2)+1
           end if
        end do
     end do
  end do

  do i=3, a_csc%n+2
     ata_csr%iptr(i) = ata_csr%iptr(i)+ata_csr%iptr(i-1)
  end do

  ata_csr%nz = ata_csr%iptr(a_csr%n+2)-1
  call qrm_alloc(ata_csr%jcn, ata_csr%nz, err)
  __QRM_INFO_CHECK(err, name, 'qrm_alloc', 9999)

  call qrm_alloc(ata_csr%val, ata_csr%nz, err)
  __QRM_INFO_CHECK(err, name, 'qrm_alloc', 9999)

  ! second pass to fill up
  mark = 0
  idx  = 0
  do col1=1, a_csc%n
     ! loop over all the columns of A
     do i=a_csc%jptr(col1), a_csc%jptr(col1+1)-1
        ! for each nnz in the column, we go through the corresponding
        ! row and count all the i,j pairs
        row1 = a_csc%irn(i)

        do j=a_csr%iptr(row1), a_csr%iptr(row1+1)-1
           col2 = a_csr%jcn(j)
           ! now, the element (col1,col2) is present in A'*A. Check if
           ! it was already found, otherwise add it.
           ! skip the diagonal
           if(col1 .gt. col2) cycle

           if(mark(col2) .lt. col1) then
              mark(col2) = col1
              idx(col2)  = ata_csr%iptr(col1+1)
              ata_csr%jcn(ata_csr%iptr(col1+1)) = col2
              ata_csr%val(ata_csr%iptr(col1+1)) = a_csc%val(i)*a_csr%val(j)
              ata_csr%iptr(col1+1) = ata_csr%iptr(col1+1)+1
              ! if(col1.eq.1 .and. col2.eq.1) then
                 ! write(*,*)'---> ',col1,col2,a_csc%val(i),a_csr%val(j)
              ! end if
           else if(mark(col2) .eq. col1) then
              ata_csr%val(idx(col2)) = ata_csr%val(idx(col2)) + a_csc%val(i)*a_csr%val(j)
              ! if(col1.eq.1 .and. col2.eq.1) then
                 ! write(*,*)'---> ',col1,col2,a_csc%val(i),a_csr%val(j)
              ! end if
           end if
        end do
     end do
  end do

  ata_csr%n = a_csc%n
  ata_csr%m = a_csc%n

  ata_csr%fmt='csr'
  ata_csr%sym=1
  
  call cqrm_spmat_dealloc(a_csr, info=err)
  call cqrm_spmat_destroy(a_csr, info=err)
  __QRM_INFO_CHECK(err, name, 'qrm_spmat_destroy', 9999)
  call qrm_dealloc(mark, err)
  call qrm_dealloc(idx , err)
  __QRM_INFO_CHECK(err, name, 'qrm_dealloc', 9999)

  if(present(info)) info = err
  return
  
9999 continue
  ! cleanup and return
  call cqrm_spmat_dealloc(a_csr)
  call cqrm_spmat_destroy(a_csr)
  call cqrm_spmat_dealloc(ata_csr)
  call cqrm_spmat_destroy(ata_csr)
  call qrm_dealloc(mark)
  call qrm_dealloc(idx)

  if(present(info)) info = err
  return

end subroutine cqrm_spmat_ata
