!! ##############################################################################################
!!
!! Copyright 2012-2024 CNRS, INPT
!! Copyright 2013-2015 UPS
!!
!! This file is part of qr_mumps.
!!
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as
!! published by the Free Software Foundation, either version 3 of
!! the License, or (at your option) any later version.
!!
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!> @file qrm_test_err.F90
!! This file contains tests for the error handling
!!
!!
!! ##############################################################################################

function zqrm_test_dense(c) result(ok)
  use zqrm_testing_mod, protect => zqrm_test_dense
  implicit none

  integer :: c
  logical :: ok

  integer, parameter :: ncases=15
  logical :: cases(ncases)

  if(c .eq. -1) then
     cases = .true.
  else if (c .le. ncases) then
     cases = .false.
     cases(c) = .true.
  end if

  ok = .true.

  if(cases(1 )) ok = case1 ()  .and. ok
  if(cases(2 )) ok = case2 ()  .and. ok
  if(cases(3 )) ok = case3 ()  .and. ok
  if(cases(4 )) ok = case4 ()  .and. ok
  if(cases(5 )) ok = case5 ()  .and. ok
  if(cases(6 )) ok = case6 ()  .and. ok
  ! if(cases(7 )) ok = case7 ()  .and. ok

  return


contains

  !> @brief Test the copy operation
  function case1() result(ok)
    use zqrm_mod
    implicit none

    logical               :: ok

    integer               :: info

    type(zqrm_dsmat_type) :: a, b
    complex(r64)             :: alpha
    real(r64)             :: anrm, bnrm

    integer               :: m, n, k, mb, mp, np, kp, i, j
    integer, allocatable  :: mbsa(:), mbsb(:)

    ok = .true.


    ! Subcase 1. simple
    m  = 12
    n  = 12
    mb = 5

    call qrm_dsmat_init(a, m, n, mb, info=info)
    call qrm_dsmat_init(b, m, n, mb, info=info)
    call qrm_dsmat_fill(a, init='r')
    call qrm_dsmat_copy(a, b, info=info)

    call qrm_dsmat_axpy(a, b, alpha=qrm_zmone, info=info)

    call qrm_dsmat_nrm(b, bnrm, info=info)
    call qrm_dsmat_destroy(a)
    call qrm_dsmat_destroy(b)

    call zqrm_prnt_testmesg(8, "dense", 1, 1, -1, (info .eq. 0) .and. (bnrm .lt. eps))

    ok = ok .and. (info .eq. 0) .and. (bnrm .lt. eps)

    ! Subcase 2. variable size tiles
    m  = 25
    n  = 25
    mbsa = (/4, 3, 6, 5, 4, 3/)
    mbsb = (/6, 5, 4, 3, 3, 4/)

    call qrm_dsmat_init(a, m, n, mbs=mbsa, info=info)
    call qrm_dsmat_init(b, m, n, mbs=mbsb, info=info)
    call qrm_dsmat_fill(a, init='r')
    call qrm_dsmat_copy(a, b, info=info)

    call qrm_dsmat_axpy(a, b, alpha=qrm_zmone, info=info)

    call qrm_dsmat_nrm(b, bnrm, info=info)
    call qrm_dsmat_destroy(a)
    call qrm_dsmat_destroy(b)

    call zqrm_prnt_testmesg(8, "dense", 1, 2, -1, (info .eq. 0) .and. (bnrm .lt. eps))

    ok = ok .and. (info .eq. 0) .and. (bnrm .lt. eps)

    return

  end function case1

  !> @brief Test the GEMM operation
  function case2() result(ok)
    use zqrm_mod
    implicit none

    logical               :: ok

    integer               :: info

    type(zqrm_dsmat_type) :: a, b, c, ctmp
    complex(r64)          :: alpha, beta, avalue, bvalue, cvalue
    real(r64)             :: areal, cnrm
    integer, allocatable  :: mbs(:)

    integer               :: m, n, k, mb, mp, np, kp

    ok = .true.

    ! Subcase 1 notrans(A), notrans(B)

    m  = 33
    n  = 27
    k  = 16
    mb = 5
    kp = 13
    mp = 16
    np = 21

#if defined(dprec) || defined(sprec)
    call random_number(avalue)
#elif defined(zprec) || defined (cprec)
    call random_number(areal)
    avalue = cmplx(areal, qrm_dzero)
#endif
    bvalue = qrm_zone / (avalue*kp)
    cvalue = qrm_zone

    call qrm_dsmat_init(a   , m,   k, mb, info=info)
    call qrm_dsmat_init(b   , k,   n, mb, info=info)
    call qrm_dsmat_init(c   , m,   n, mb, info=info)
    call qrm_dsmat_init(ctmp, mp, np, mb, info=info)
    call qrm_dsmat_fill(a,    init='v', ivalue=avalue)
    call qrm_dsmat_fill(b,    init='v', ivalue=bvalue)
    call qrm_dsmat_fill(c,    init='v', ivalue=qrm_zzero)
    call qrm_dsmat_fill(ctmp, init='v', ivalue=cvalue)

    call qrm_dsmat_copy(ctmp, c, info=info)

    call qrm_dsmat_gemm(qrm_no_transp, qrm_no_transp, qrm_zmone, a, b, qrm_zone, c, m=mp, n=np, k=kp)

    call qrm_dsmat_nrm(c, cnrm, info=info)
    call qrm_dsmat_destroy(a)
    call qrm_dsmat_destroy(b)
    call qrm_dsmat_destroy(c)
    call qrm_dsmat_destroy(ctmp)

    call zqrm_prnt_testmesg(8, "dense", 2, 1, -1, (info .eq. 0) .and. (cnrm .lt. eps))

    ok = ok .and. (info .eq. 0) .and. (cnrm .lt. eps)


    ! Subcase 2 trans(A), notrans(B)

#if defined(dprec) || defined(sprec)
    call random_number(avalue)
#elif defined(zprec) || defined (cprec)
    call random_number(areal)
    avalue = cmplx(areal, qrm_dzero)
#endif
    bvalue = qrm_zone / (avalue*kp)
    cvalue = qrm_zone

    call qrm_dsmat_init(a   , k , m , mb, info=info)
    call qrm_dsmat_init(b   , k , n , mb, info=info)
    call qrm_dsmat_init(c   , m , n , mb, info=info)
    call qrm_dsmat_init(ctmp, mp, np, mb, info=info)

    call qrm_dsmat_fill(a,    init='v', ivalue=avalue)
    call qrm_dsmat_fill(b,    init='v', ivalue=bvalue)
    call qrm_dsmat_fill(c,    init='v', ivalue=qrm_zzero)
    call qrm_dsmat_fill(ctmp, init='v', ivalue=cvalue)
    call qrm_dsmat_copy(ctmp, c, info=info)

    call qrm_dsmat_gemm(qrm_conj_transp, qrm_no_transp, qrm_zmone, a, b, qrm_zone, c, m=mp, n=np, k=kp, info=info)

    call qrm_dsmat_nrm(c, cnrm, info=info)
    call qrm_dsmat_destroy(a)
    call qrm_dsmat_destroy(b)
    call qrm_dsmat_destroy(c)
    call qrm_dsmat_destroy(ctmp)

    call zqrm_prnt_testmesg(8, "dense", 2, 2, -1, (info .eq. 0) .and. (cnrm .lt. eps))

    ok = ok .and. (info .eq. 0) .and. (cnrm .lt. eps)


    ! Subcase 3 trans(A), trans(B)

#if defined(dprec) || defined(sprec)
    call random_number(avalue)
#elif defined(zprec) || defined (cprec)
    call random_number(areal)
    avalue = cmplx(areal, qrm_dzero)
#endif
    bvalue = qrm_zone / (avalue*kp)
    cvalue = qrm_zone

    call qrm_dsmat_init(a   ,  k,  m, mb, info=info)
    call qrm_dsmat_init(b   ,  n,  k, mb, info=info)
    call qrm_dsmat_init(c   ,  m,  n, mb, info=info)
    call qrm_dsmat_init(ctmp, mp, np, mb, info=info)
    call qrm_dsmat_fill(a,    init='v', ivalue=avalue)
    call qrm_dsmat_fill(b,    init='v', ivalue=bvalue)
    call qrm_dsmat_fill(c,    init='v', ivalue=qrm_zzero)
    call qrm_dsmat_fill(ctmp, init='v', ivalue=cvalue)

    call qrm_dsmat_copy(ctmp, c, info=info)

    call qrm_dsmat_gemm(qrm_conj_transp, qrm_conj_transp, qrm_zmone, a, b, qrm_zone, c, m=mp, n=np, k=kp, info=info)

    call qrm_dsmat_nrm(c, cnrm, info=info)
    call qrm_dsmat_destroy(a)
    call qrm_dsmat_destroy(b)
    call qrm_dsmat_destroy(c)
    call qrm_dsmat_destroy(ctmp)

    call zqrm_prnt_testmesg(8, "dense", 2, 3, -1, (info .eq. 0) .and. (cnrm .lt. eps))

    ok = ok .and. (info .eq. 0) .and. (cnrm .lt. eps)

    
    ! Subcase 4 trans(A), trans(B) with variable size tiles

#if defined(dprec) || defined(sprec)
    call random_number(avalue)
#elif defined(zprec) || defined (cprec)
    call random_number(areal)
    avalue = cmplx(areal, qrm_dzero)
#endif
    bvalue = qrm_zone / (avalue*kp)
    cvalue = qrm_zone

    mbs = (/3, 5, 4, 6, 4, 5, 6, 3/)
    
    call qrm_dsmat_init(a   , m,   k, mbs=mbs, info=info)
    call qrm_dsmat_init(b   , k,   n, mbs=mbs, info=info)
    call qrm_dsmat_init(c   , m,   n, mbs=mbs, info=info)
    call qrm_dsmat_init(ctmp, mp, np, mbs=mbs, info=info)
    call qrm_dsmat_fill(a,    init='v', ivalue=avalue)
    call qrm_dsmat_fill(b,    init='v', ivalue=bvalue)
    call qrm_dsmat_fill(c,    init='v', ivalue=qrm_zzero)
    call qrm_dsmat_fill(ctmp, init='v', ivalue=cvalue)

    call qrm_dsmat_copy(ctmp, c, info=info)

    call qrm_dsmat_gemm(qrm_no_transp, qrm_no_transp, qrm_zmone, a, b, qrm_zone, c, m=mp, n=np, k=kp)

    call qrm_dsmat_nrm(c, cnrm, info=info)
    call qrm_dsmat_destroy(a)
    call qrm_dsmat_destroy(b)
    call qrm_dsmat_destroy(c)
    call qrm_dsmat_destroy(ctmp)

    call zqrm_prnt_testmesg(8, "dense", 2, 4, -1, (info .eq. 0) .and. (cnrm .lt. eps))

    ok = ok .and. (info .eq. 0) .and. (cnrm .lt. eps)


    ! Subcase 5 trans(A), notrans(B) with variable size tiles

#if defined(dprec) || defined(sprec)
    call random_number(avalue)
#elif defined(zprec) || defined (cprec)
    call random_number(areal)
    avalue = cmplx(areal, qrm_dzero)
#endif
    bvalue = qrm_zone / (avalue*kp)
    cvalue = qrm_zone

    call qrm_dsmat_init(a   , k , m , mbs=mbs, info=info)
    call qrm_dsmat_init(b   , k , n , mbs=mbs, info=info)
    call qrm_dsmat_init(c   , m , n , mbs=mbs, info=info)
    call qrm_dsmat_init(ctmp, mp, np, mbs=mbs, info=info)

    call qrm_dsmat_fill(a,    init='v', ivalue=avalue)
    call qrm_dsmat_fill(b,    init='v', ivalue=bvalue)
    call qrm_dsmat_fill(c,    init='v', ivalue=qrm_zzero)
    call qrm_dsmat_fill(ctmp, init='v', ivalue=cvalue)
    call qrm_dsmat_copy(ctmp, c, info=info)

    call qrm_dsmat_gemm(qrm_conj_transp, qrm_no_transp, qrm_zmone, a, b, qrm_zone, c, m=mp, n=np, k=kp, info=info)

    call qrm_dsmat_nrm(c, cnrm, info=info)
    call qrm_dsmat_destroy(a)
    call qrm_dsmat_destroy(b)
    call qrm_dsmat_destroy(c)
    call qrm_dsmat_destroy(ctmp)

    call zqrm_prnt_testmesg(8, "dense", 2, 5, -1, (info .eq. 0) .and. (cnrm .lt. eps))

    ok = ok .and. (info .eq. 0) .and. (cnrm .lt. eps)


    ! Subcase 6 trans(A), trans(B) with variable size tiles

#if defined(dprec) || defined(sprec)
    call random_number(avalue)
#elif defined(zprec) || defined (cprec)
    call random_number(areal)
    avalue = cmplx(areal, qrm_dzero)
#endif
    bvalue = qrm_zone / (avalue*kp)
    cvalue = qrm_zone

    call qrm_dsmat_init(a   ,  k,  m, mbs=mbs, info=info)
    call qrm_dsmat_init(b   ,  n,  k, mbs=mbs, info=info)
    call qrm_dsmat_init(c   ,  m,  n, mbs=mbs, info=info)
    call qrm_dsmat_init(ctmp, mp, np, mbs=mbs, info=info)
    call qrm_dsmat_fill(a,    init='v', ivalue=avalue)
    call qrm_dsmat_fill(b,    init='v', ivalue=bvalue)
    call qrm_dsmat_fill(c,    init='v', ivalue=qrm_zzero)
    call qrm_dsmat_fill(ctmp, init='v', ivalue=cvalue)

    call qrm_dsmat_copy(ctmp, c, info=info)

    call qrm_dsmat_gemm(qrm_conj_transp, qrm_conj_transp, qrm_zmone, a, b, qrm_zone, c, m=mp, n=np, k=kp, info=info)

    call qrm_dsmat_nrm(c, cnrm, info=info)
    call qrm_dsmat_destroy(a)
    call qrm_dsmat_destroy(b)
    call qrm_dsmat_destroy(c)
    call qrm_dsmat_destroy(ctmp)

    call zqrm_prnt_testmesg(8, "dense", 2, 6, -1, (info .eq. 0) .and. (cnrm .lt. eps))

    ok = ok .and. (info .eq. 0) .and. (cnrm .lt. eps)

    
    return

  end function case2




  !> @brief Test the TRSM and TRMM operations
  function case3() result(ok)
    use zqrm_mod
    implicit none

    logical                :: ok

    integer                :: info

    type(zqrm_dsmat_type)  :: a, b, x
    complex(r64)              :: alpha, beta, avalue, bvalue, cvalue
    real(r64)              :: bnrm
    complex(r64), allocatable :: ad(:,:)
    integer, allocatable   :: mbs(:)

    integer                :: m, n, k, mb, mp, np, kp, i

    ok = .true.

    ! Subcase 1 notrans(A)

    m  = 19
    n  = 14
    k  = 27
    mb = 4
    mp = 15
    np = 13
    kp = 23

    call qrm_dsmat_init(a   , m, k, mb, info=info)
    call qrm_dsmat_init(b   , m, n, mb, info=info)
    call qrm_dsmat_init(x   , k, n, mb, info=info)
    call qrm_dsmat_fill(a, init='r')
    call qrm_dsmat_fill(b, init='r')
    call qrm_dsmat_fill(x, init='r')
    call qrm_dsmat_copy(b, x, m=mp, info=info)

    ! make a diag-dominant
    call qrm_dsmat_addi(a, v=qrm_ztwo)
    
    call qrm_dsmat_trsm(qrm_left, qrm_upper, qrm_no_transp, qrm_no_unit, qrm_zone, a, x, m=mp, n=np, k=kp)

    call qrm_dsmat_trmm(qrm_left, qrm_upper, qrm_no_transp, qrm_no_unit, qrm_zone, a, x, m=mp, n=np, k=kp, info=info)

    call qrm_dsmat_axpy(x, b, m=mp, n=np, alpha=qrm_zmone, info=info)
    call qrm_dsmat_nrm(b, bnrm, m=mp, n=np, info=info)

    call qrm_dsmat_destroy(a)
    call qrm_dsmat_destroy(b)
    call qrm_dsmat_destroy(x)

    call zqrm_prnt_testmesg(8, "dense", 3, 1, -1, (info .eq. 0) .and. (bnrm .lt. eps))

    ok = ok .and. (info .eq. 0) .and. (bnrm .lt. eps)

    ! Subcase 2 trans(A)
    m  = 19
    n  = 14
    k  = 27
    mb = 4
    mp = 15
    np = 13
    kp = 23

    call qrm_dsmat_init(a,  m,  k, mb, info=info)
    call qrm_dsmat_init(b, kp, np, mb, info=info)
    call qrm_dsmat_init(x, kp, np, mb, info=info)
    call qrm_dsmat_fill(a, init='r')
    call qrm_dsmat_fill(b, init='r')
    call qrm_dsmat_copy(b, x, info=info)

    ! make a diag-dominant
    call qrm_dsmat_addi(a, qrm_ztwo)

    call qrm_dsmat_trsm(qrm_left, qrm_upper, qrm_conj_transp, qrm_no_unit, &
         qrm_zone, a, x, m=mp, n=np, k=kp, info=info)

    call qrm_dsmat_trmm(qrm_left, qrm_upper, qrm_conj_transp, qrm_no_unit, &
         qrm_zone, a, x, m=mp, n=np, k=kp, info=info)
    
    call qrm_dsmat_axpy(x, b, alpha=qrm_zmone, m=kp, n=np, info=info)
    
    call qrm_dsmat_nrm(b, bnrm, info=info)

    call qrm_dsmat_destroy(a)
    call qrm_dsmat_destroy(b)
    call qrm_dsmat_destroy(x)

    call zqrm_prnt_testmesg(8, "dense", 3, 2, -1, (info .eq. 0) .and. (bnrm .lt. eps))

    ok = ok .and. (info .eq. 0) .and. (bnrm .lt. eps)

    ! Subcase 3 notrans(A) with variable size tiles

    m   = 19
    n   = 14
    k   = 27
    mbs = (/3, 4, 5, 4, 6, 7/)
    mp  = 15
    np  = 13
    kp  = 23

    call qrm_dsmat_init(a   , m, k, mbs=mbs, info=info)
    call qrm_dsmat_init(b   , m, n, mbs=mbs, info=info)
    call qrm_dsmat_init(x   , k, n, mbs=mbs, info=info)
    call qrm_dsmat_fill(a, init='r')
    call qrm_dsmat_fill(b, init='r')
    call qrm_dsmat_fill(x, init='r')
    call qrm_dsmat_copy(b, x, m=mp, info=info)

    ! make a diag-dominant
    call qrm_dsmat_addi(a, v=qrm_ztwo)
    
    call qrm_dsmat_trsm(qrm_left, qrm_upper, qrm_no_transp, qrm_no_unit, qrm_zone, a, x, m=mp, n=np, k=kp)

    call qrm_dsmat_trmm(qrm_left, qrm_upper, qrm_no_transp, qrm_no_unit, qrm_zone, a, x, m=mp, n=np, k=kp, info=info)

    call qrm_dsmat_axpy(x, b, m=mp, n=np, alpha=qrm_zmone, info=info)
    call qrm_dsmat_nrm(b, bnrm, m=mp, n=np, info=info)

    call qrm_dsmat_destroy(a)
    call qrm_dsmat_destroy(b)
    call qrm_dsmat_destroy(x)

    call zqrm_prnt_testmesg(8, "dense", 3, 3, -1, (info .eq. 0) .and. (bnrm .lt. eps))

    ok = ok .and. (info .eq. 0) .and. (bnrm .lt. eps)

    ! Subcase 4 trans(A) with variable size tiles
    m  = 19
    n  = 14
    k  = 27
    mbs = (/3, 4, 5, 4, 6, 7/)
    mp = 15
    np = 13
    kp = 23

    call qrm_dsmat_init(a,  m,  k, mbs=mbs, info=info)
    call qrm_dsmat_init(b, kp, np, mbs=mbs, info=info)
    call qrm_dsmat_init(x, kp, np, mbs=mbs, info=info)
    call qrm_dsmat_fill(a, init='r')
    call qrm_dsmat_fill(b, init='r')
    call qrm_dsmat_copy(b, x, info=info)

    ! make a diag-dominant
    call qrm_dsmat_addi(a, qrm_ztwo)

    call qrm_dsmat_trsm(qrm_left, qrm_upper, qrm_conj_transp, qrm_no_unit, &
         qrm_zone, a, x, m=mp, n=np, k=kp, info=info)

    call qrm_dsmat_trmm(qrm_left, qrm_upper, qrm_conj_transp, qrm_no_unit, &
         qrm_zone, a, x, m=mp, n=np, k=kp, info=info)
    
    call qrm_dsmat_axpy(x, b, alpha=qrm_zmone, m=kp, n=np, info=info)
    
    call qrm_dsmat_nrm(b, bnrm, info=info)

    call qrm_dsmat_destroy(a)
    call qrm_dsmat_destroy(b)
    call qrm_dsmat_destroy(x)

    call zqrm_prnt_testmesg(8, "dense", 3, 4, -1, (info .eq. 0) .and. (bnrm .lt. eps))

    ok = ok .and. (info .eq. 0) .and. (bnrm .lt. eps)



    return

  end function case3




    !> @brief Test the GEQR and GEMQR operations
  function case4() result(ok)
    use zqrm_mod
    implicit none

    logical                :: ok

    integer                :: info

    type(zqrm_dsmat_type)  :: a, b, x, r, qr, t
    real(r64)              :: bnrm, anrm, onrm, xnrm, rnrm, nrm
    type(zqrm_ws_type)     :: w

    integer                :: ms(3), ns(3), mbs(2), ibs(2), bhs(3)
    character              :: ts(3), tr
    integer                :: m, n, k, mb, nb, mp, np, i, j, qm, qn, nrhs, sc, bh, ib
    integer, allocatable   :: bss(:)
    
    ok = .true.
    info = 0

    ms   = (/1200, 1000, 1000/)
    ns   = (/1000, 1200, 1200/)
    ts   = (/qrm_no_transp, qrm_no_transp, qrm_conj_transp/)
    mbs  = (/128, 128/)
    ibs  = (/32, 32/)
    bhs  = (/1110, 1, 2/)
    
    nrhs = 1

    sc = 1

    do i = 1, 3
       m  = ms(i)
       n  = ns(i)
       tr = ts(i)

       ! initialize original data
       if(tr.eq.qrm_conj_transp) then
          qm = n
          qn = m
       else
          qm = m
          qn = n
       end if

       do j = 1, 2
          mb = mbs(j)
          ib = ibs(j)

          if(j.eq.2) then
             ! test with variable size tiles
             bss = (/128, 256, 190, 210, 320, 171/)
             call qrm_dsmat_init(a , m , n , mbs=bss, info=info)
             call qrm_dsmat_init(qr, qm, qn, mbs=bss, info=info)
          else
             call qrm_dsmat_init(a , m , n , mb, info=info)
             call qrm_dsmat_init(qr, qm, qn, mb, info=info)
          end if
          
          call qrm_dsmat_fill(a , init='r')
          call qrm_dsmat_fill(qr, init='v', ivalue=qrm_zzero)

          call qrm_dsmat_nrm(a, anrm, info=info)


          do k = 1, 3
             bh = bhs(k)

             call qrm_dsmat_copy(a, qr, transp=tr, info=info)
             call qrm_dsmat_init_t(qr, t, ib, bh, info=info)
             if(j.eq.2) then
                call qrm_ws_init(w, 2*maxval(bss), ib)
             else
                call qrm_ws_init(w, 2*mb, ib)
             end if

             call qrm_dsmat_geqr(qr, t, ib, bh, w, info=info)

             if(j.eq.2) then
                call qrm_dsmat_init(b, m, nrhs, mbs=bss, info=info)
                call qrm_dsmat_init(x, n, nrhs, mbs=bss, info=info)
                call qrm_dsmat_init(r, m, nrhs, mbs=bss, info=info)
             else
                call qrm_dsmat_init(b, m, nrhs, mb, info=info)
                call qrm_dsmat_init(x, n, nrhs, mb, info=info)
                call qrm_dsmat_init(r, m, nrhs, mb, info=info)
             end if
             call qrm_dsmat_fill(b, init='r')
             call qrm_dsmat_fill(x, init='v', ivalue=qrm_zzero)
             call qrm_dsmat_fill(r, init='v', ivalue=qrm_zzero)
             call qrm_dsmat_copy(b, r, info=info)
             call qrm_dsmat_nrm(b, bnrm, info=info)

             if((m.gt.n) .or. (tr.ne.qrm_conj_transp)) then
                call qrm_dsmat_gemqr(qrm_conj_transp, qr, t, b, ib, bh, w, info=info)
                call qrm_dsmat_copy(b, x, m=min(m,n), info=info)
                call qrm_dsmat_trsm(qrm_left, qrm_upper, qrm_no_transp, qrm_no_unit, qrm_zone, qr, x, info=info)
             else
                call qrm_dsmat_trsm(qrm_left, qrm_upper, qrm_conj_transp, qrm_no_unit, qrm_zone, qr, b, info=info)
                call qrm_dsmat_copy(b, x, m=min(m,n), info=info)
                call qrm_dsmat_gemqr(qrm_no_transp, qr, t, x, ib, bh, w, info=info)
             end if

             ! one more copy to make sure bfull is not modified
             call qrm_dsmat_nrm(x, xnrm, info=info)
             call qrm_dsmat_gemm(qrm_no_transp, qrm_no_transp, qrm_zmone, a, x, qrm_zone, r)
             ! compute A'*r and store the result in x
             call qrm_dsmat_gemm(qrm_conj_transp, qrm_no_transp, qrm_zone, a, r, qrm_zzero, x)
             call qrm_dsmat_nrm(x, onrm, info=info)
             call qrm_dsmat_nrm(r, rnrm, info=info)

             if(m.gt.n) then
                nrm = onrm/anrm
             else
                nrm = rnrm/anrm
             end if

             call zqrm_prnt_testmesg(8, "dense", 4, sc, -1, (info .eq. 0) .and. (nrm .lt. eps))
             ok = ok .and. (info .eq. 0) .and. (nrm .lt. eps)

             call qrm_dsmat_destroy(t)
             call qrm_dsmat_destroy(b)
             call qrm_dsmat_destroy(x)
             call qrm_dsmat_destroy(r)
             call qrm_ws_destroy(w)

             sc = sc+1
             
          end do

          call qrm_dsmat_destroy(a)
          call qrm_dsmat_destroy(qr)

       end do
    end do


    return

  end function case4


  !> @brief Test the TRDCN operation
  function case5() result(ok)
    use zqrm_mod
    implicit none

    logical                :: ok

    integer                :: info

    type(zqrm_dsmat_type)  :: a, t
    real(r64)              :: anrm
    type(zqrm_ws_type)     :: w

    character              :: ts(3), tr
    integer                :: m, n, mb, bh, ib, d

    ok = .true.

    info = 0

    m  = 150
    n  = 100
    mb = 32
    ib = 16
    bh = 1000
    d  = 0
    
    call qrm_dsmat_init(a , m , n , mb, info=info)
    call qrm_dsmat_fill(a, init='r')
    ! make a rank deficient
    call qrm_dsmat_copy(a, a, ja=30, jb=90, m=m, n=1)
    ! call qrm_dsmat_print(a,'f5.2')
    
    call qrm_dsmat_nrm(a, anrm, info=info)
    call qrm_dsmat_init_t(a, t, ib, bh, info=info)
    call qrm_ws_init(w, 2*mb, ib)

    call qrm_dsmat_geqr(a, t, ib, bh, w, info=info)
    call qrm_dsmat_trdcn(a, d, eps*anrm, n, info)

    call zqrm_prnt_testmesg(8, "dense", 5, 1, -1, (info .eq. 0) .and. (d .gt. 0))
    ok = ok .and. (info .eq. 0) .and. (d .gt. 0)

    call qrm_dsmat_destroy(t)
    call qrm_ws_destroy(w)

    call qrm_dsmat_destroy(a)


    return

  end function case5




    !> @brief Test the POTRF 
  function case6() result(ok)
    use zqrm_mod
    implicit none

    logical                :: ok

    integer                :: info

    type(zqrm_dsmat_type)  :: a, b, x, r, rr
    real(r64)              :: bnrm, anrm, xnrm, rnrm, nrm

    integer                :: ms(3), mbs(2)
    integer                :: m, mb, i, j, nrhs, sc, ii, jj
    integer, allocatable   :: bss(:)
    
    ok = .true.
    info = 0

    ms   = (/1093, 1000, 927/)
    mbs  = (/128, 113/)
    
    nrhs = 1

    sc = 1

    do i = 1, 3
       m  = ms(i)

       do j = 1, 3
          if(j.eq.3) then
             mb  = -1
             bss = (/215, 128, 234, 197, 212, 154/)
             call qrm_dsmat_init(a , m, m, mbs=bss, info=info)
             call qrm_dsmat_init(rr, m, m, mbs=bss, info=info)
          else
             mb = mbs(j)
             call qrm_dsmat_init(a , m, m, mb=mb, info=info)
             call qrm_dsmat_init(rr, m, m, mb=mb, info=info)
          end if
                    
          ! buils a hermitian spd  matrix
          call qrm_dsmat_fill(rr, 'r')
          call qrm_dsmat_gemm(qrm_conj_transp, qrm_no_transp,    &
               qrm_zone,                                         &
               rr, rr,                                           &
               qrm_zzero,                                        &
               a)
          call qrm_dsmat_copy(a, rr, info=info)
          call qrm_dsmat_nrm(a, anrm, info=info)

          call qrm_dsmat_potrf(qrm_upper, rr, info=info)

          if(j.eq.3) then
             call qrm_dsmat_init(b , m, nrhs, mbs=bss, info=info)
             call qrm_dsmat_init(r , m, nrhs, mbs=bss, info=info)
             call qrm_dsmat_init(x , m, nrhs, mbs=bss, info=info)
          else
             call qrm_dsmat_init(b , m, nrhs, mb, info=info)
             call qrm_dsmat_init(r , m, nrhs, mb, info=info)
             call qrm_dsmat_init(x , m, nrhs, mb, info=info)
          end if

          call qrm_dsmat_fill(b, init='v', ivalue=qrm_zone)
          call qrm_dsmat_fill(r, init='v', ivalue=qrm_zzero)
          call qrm_dsmat_fill(x, init='v', ivalue=qrm_zzero)

          call qrm_dsmat_copy(b, x, info=info)

          call qrm_dsmat_nrm(b, bnrm, info=info)

          call qrm_dsmat_trsm(qrm_left, qrm_upper, &
               qrm_conj_transp, qrm_no_unit,       &
               qrm_zone, rr, x, info=info)
          call qrm_dsmat_trsm(qrm_left, qrm_upper, &
               qrm_no_transp, qrm_no_unit,         &
               qrm_zone, rr, x, info=info)

          call qrm_dsmat_nrm(x, xnrm, info=info)
          call qrm_dsmat_copy(b, r, info=info)
          call qrm_dsmat_gemm(qrm_no_transp, qrm_no_transp, qrm_zone, a, x, qrm_zmone, r, info=info)
          call qrm_dsmat_nrm(r, rnrm, info=info)
          
          nrm = rnrm/(anrm*bnrm+xnrm)

          call zqrm_prnt_testmesg(8, "dense", 6, sc, -1, (info .eq. 0) .and. (nrm .lt. eps))
          ok = ok .and. (info .eq. 0) .and. (nrm .lt. eps)
          
          call qrm_dsmat_destroy(b)
          call qrm_dsmat_destroy(x)
          call qrm_dsmat_destroy(r)

          call qrm_dsmat_destroy(a)
          call qrm_dsmat_destroy(rr)
          sc = sc+1
       end do

    end do

    return

  end function case6



    !> @brief Test the GEQR and GEMQR operations
  ! function case7() result(ok)
  !   use zqrm_mod
  !   implicit none

  !   logical                :: ok

  !   integer                :: info

  !   type(zqrm_dsmat_type)  :: a, c, ba, bc, ra, rc, x, qra, qrc, t, tmpa, tmpc
  !   real(r64)              :: banrm, bcnrm, bnrm, onrm, xnrm, rnrm, nrm
  !   type(zqrm_ws_type)     :: w

  !   integer                :: ms(3), ns(3), mbs(2), ibs(2)
  !   character              :: tss(2), ts
  !   integer                :: m, n, k, l, mb, mp, np, i, j, qm, qn, nrhs, sc, ii, jj, ib
  !   integer, parameter     :: izero=0
    
  !   ok = .true.
  !   info = 0

  !   ms   = (/200, 180, 220/)
  !   ns   = (/200, 200, 200/)
  !   tss  = (/qrm_square, qrm_triang/)
  !   mbs  = (/32, 40/)
  !   ibs  = (/8, 16/)

  !   ! ms   = (/30, 35, 27/)
  !   ! ns   = (/30, 47, 25/)
  !   ! tss  = (/'s', 't'/)
  !   ! mbs  = (/32, 6/)
  !   ! ibs  = (/32, 2/)
    
  !   nrhs = 1

  !   sc = 1

  !   do i = 1, 3
  !      m  = ms(i)
  !      n  = ns(i)

  !      do j = 1, 2
  !         mb = mbs(j)
  !         ib = ibs(j)

  !         call qrm_dsmat_init(a , n, n, mb, info=info)
  !         call qrm_dsmat_init(c , m, n, mb, info=info)
  !         call qrm_dsmat_fill(a, init='r')
  !         call qrm_dsmat_fill(c, init='r')
          
  !         ! make [A; C] well conditioned. TODO: implement a proper way to do this
  !         do ii=1, size(a%blocks,1)
  !            do jj=1, size(a%blocks,2)
  !               do k=1,min(size(a%blocks(ii,jj)%c,1),size(a%blocks(ii,jj)%c,2))
  !                  a%blocks(ii,jj)%c(k,k) = a%blocks(ii,jj)%c(k,k) + 10*m
  !               end do
  !            end do
  !         end do

  !         do k = 1, 2
  !            ts = tss(k)
  !            if(ts.eq.qrm_square) then
  !               l = 0
  !            else
  !               l = m
  !            end if

  !            call qrm_dsmat_init(qra, n, n, mb, init='v', ivalue=qrm_zzero, info=info)
  !            call qrm_dsmat_init(qrc, m, n, mb, init='v', ivalue=qrm_zzero, info=info)
  !            call qrm_dsmat_copy(a, qra, info=info)
  !            call qrm_dsmat_copy(c, qrc, info=info)
       
  !            call qrm_dsmat_init_t(qrc, t, ib, 0, l=l, info=info)
  !            call qrm_ws_init(w, 2*mb, ib)

  !            call qrm_dsmat_tpqr(qra, qrc, t, ib, w, l=l, info=info)

  !            call qrm_dsmat_init(ba, n, 1, mb, init='r', info=info)
  !            call qrm_dsmat_init(bc, m, 1, mb, init='r', info=info)
             
  !            call qrm_dsmat_nrm(ba, banrm, info=info)
  !            call qrm_dsmat_nrm(bc, bcnrm, info=info)
  !            bnrm = sqrt(banrm**2+bcnrm**2)
             
  !            call qrm_dsmat_init(x , n, 1, mb, init='v', ivalue=qrm_zzero, info=info)

  !            call qrm_dsmat_init(ra, n, 1, mb, init='v', ivalue=qrm_zzero, info=info)
  !            call qrm_dsmat_init(rc, m, 1, mb, init='v', ivalue=qrm_zzero, info=info)
  !            call qrm_dsmat_copy(ba, ra, info=info)
  !            call qrm_dsmat_copy(bc, rc, info=info)

  !            ! call qrm_dsmat_print(qra,'f25.15', triu=.true., unit=22, zero=.true.)

  !            call qrm_dsmat_tpmqr(qrc, ba, bc, t, ib, w, l=l, info=info)
  !            call qrm_dsmat_copy(ba, x, info=info)
  !            call qrm_dsmat_trsm(qrm_left, qrm_upper, qrm_no_transp, qrm_no_unit, &
  !                 qrm_zone, qra, x, info=info)
  !            call qrm_dsmat_nrm(x, xnrm, info=info)

  !            ! compute residual
  !            call qrm_dsmat_init(tmpa, n, 1, mb, init='v', ivalue=qrm_zzero, info=info)
  !            call qrm_dsmat_copy(x, tmpa, info=info)
  !            call qrm_dsmat_trmm(qrm_left, qrm_upper, &
  !                 qrm_no_transp, qrm_no_unit,         &
  !                 qrm_zone, a, tmpa, n, 1, n, info=info)
  !            call qrm_dsmat_axpy(tmpa, ra, alpha=qrm_zmone, info=info)
  !            call qrm_dsmat_destroy(tmpa)

             
  !            if(ts.eq.qrm_triang) then
  !               call qrm_dsmat_init(tmpc, n, 1, mb, init='v', ivalue=qrm_zzero, info=info)
  !               call qrm_dsmat_copy(x, tmpc, m=n, info=info)
  !               call qrm_dsmat_trmm(qrm_left, qrm_upper, &
  !                    qrm_no_transp, qrm_no_unit,         &
  !                    qrm_zone, c, tmpc, m=min(m,n), n=1, k=n, info=info)
  !               call qrm_dsmat_axpy(tmpc, rc, m=min(m,n), alpha=qrm_zmone, info=info)
  !               call qrm_dsmat_destroy(tmpc)
  !            else
  !               call qrm_dsmat_gemm(qrm_no_transp, qrm_no_transp, qrm_zmone, c, x, qrm_zone, rc)
  !            end if

  !            call qrm_dsmat_trmm(qrm_left, qrm_upper, &
  !                 qrm_conj_transp, qrm_no_unit,       &
  !                 qrm_zone, a, ra, n, 1, n, info=info)
  !            if(ts.eq.qrm_triang) then
  !               call qrm_dsmat_init(tmpc, n, 1, mb, init='v', ivalue=qrm_zzero, info=info)
  !               call qrm_dsmat_copy(rc, tmpc, m=min(m,n), info=info)
  !               call qrm_dsmat_trmm(qrm_left, qrm_upper,  &
  !                    qrm_conj_transp, qrm_no_unit,        &
  !                    qrm_zone, c, tmpc, m=min(m,n), n=1, k=n, info=info)
  !               call qrm_dsmat_axpy(tmpc, ra, m=n, alpha=qrm_zone, info=info)
  !               call qrm_dsmat_destroy(tmpc)
  !            else
  !               call qrm_dsmat_gemm(qrm_conj_transp, qrm_no_transp, qrm_zone, c, rc, qrm_zone, ra)
  !            end if

  !            ! call qrm_dsmat_print(ra,'es15.7')
  !            call qrm_dsmat_nrm(ra, onrm, info=info)

  !            nrm = onrm/bnrm
  !            ! write(*,*)'---> ',nrm
             
  !            call zqrm_prnt_testmesg(8, "dense", 7, sc, -1, (info .eq. 0) .and. (nrm .lt. eps))
  !            ok = ok .and. (info .eq. 0) .and. (nrm .lt. eps)
             
  !            call qrm_dsmat_destroy(qra)
  !            call qrm_dsmat_destroy(qrc)
  !            call qrm_dsmat_destroy(ra)
  !            call qrm_dsmat_destroy(rc)
  !            call qrm_dsmat_destroy(t)
  !            call qrm_dsmat_destroy(ba)
  !            call qrm_dsmat_destroy(bc)
  !            call qrm_dsmat_destroy(x)
  !            call qrm_ws_destroy(w)

  !            sc = sc+1
             
  !         end do

  !         call qrm_dsmat_destroy(a)
  !         call qrm_dsmat_destroy(c)

  !      end do
  !   end do


  !   return

  ! end function case7


  

end function zqrm_test_dense
