!! ##############################################################################################
!!
!! Copyright 2012-2024 CNRS, INPT
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################

program cqrm_dsgemm

  use cqrm_dsmat_mod
  use cqrm_mod
  use qrm_string_mod
#if defined(have_starpu) 
  use fstarpu_mod
#endif
  implicit none


  type(cqrm_dsmat_type), target  :: a, b, c
  type(qrm_dscr_type)            :: qrm_dscr

  integer                        :: info, verb, ncpu, ngpu
  integer                        :: am, an, bm, bn, cm, cn, m, n, k, mb, i
  character                      :: transa, transb
  integer(kind=8)                :: ts, te, tr
  real(kind(1.d0))               :: tgemm, gflops
  logical                        :: dopin=.true.
  
  call qrm_init()
  call qrm_stop_profiling()

  call get_cla(m, n, k, mb, transa, transb)

  call qrm_dscr_init(qrm_dscr)

  call qrm_get('qrm_ncpu', ncpu)
  call qrm_get('qrm_ngpu', ngpu)
  call qrm_get('qrm_mb'  , mb)

  call qrm_log_record("Setup", verb=2)
  call qrm_log_record("- M","m","i10",m)
  call qrm_log_record("- N","n","i10",n)
  call qrm_log_record("- MB","mb","i4",mb)
  call qrm_log_record("- NCPU","ncpu","i4",ncpu)
  call qrm_log_record("- NGPU","ngpu","i4",ngpu)
  
  cm = m
  cn = n
  if(transa.eq.qrm_nt_transp .or. transa.eq.qrm_nt_conj_transp) then
     am = k
     an = m
  else
     am = m
     an = k
  end if

  if(transb.eq.qrm_nt_transp .or. transb.eq.qrm_nt_conj_transp) then
     bm = m
     bn = k
  else
     bm = k
     bn = n
  end if

  gflops = 2*(real(m)*real(n)*real(k))/1e9

  call qrm_dsmat_init(a, am, an, mb=mb, pin=dopin, info=info)
  call qrm_dsmat_init(b, bm, bn, mb=mb, pin=dopin, info=info)
  call qrm_dsmat_init(c, cm, cn, mb=mb, pin=dopin, info=info)
  
  call qrm_dsmat_fill(a, 'r')
  call qrm_dsmat_fill(b, 'r')
  call qrm_dsmat_fill(c, 'r')

  call qrm_start_profiling()
  call system_clock(ts)
  call qrm_dsmat_gemm(transa, transb, &
                      qrm_cone,       &
                      a, b,           &
                      qrm_cone,       &
                      c)
  call system_clock(te,tr)
  call qrm_stop_profiling()
  tgemm = max(real(te-ts)/real(tr),1e-20) ! the max is to avoid signaling divide by zero

  call qrm_log_record("GEMM", verb=2)
  call qrm_log_record("- Time for GEMM","time","es10.3",tgemm)
  call qrm_log_record("- Gflops/s","gflops","f8.2",gflops/tgemm)
  call qrm_log_record("=", verb=2)
  
10 continue 

  call qrm_log_print('list',verb=2)
  ! cleanup
  
  call qrm_dsmat_destroy(a)
  call qrm_dsmat_destroy(b)
  call qrm_dsmat_destroy(c)
     
  call qrm_finalize()
     
  stop


contains


  subroutine get_cla(m, n, k, mb, transa, transb)
    use qrm_cli_mod
    implicit none

    integer            :: m, n, k, mb
    type(qrm_cli_type) :: cli
    character          :: transa, transb
    character(len=:), allocatable :: tra, trb
    integer            :: eunit

    call qrm_get('qrm_eunit', eunit)

    call cli%init(title='cqrm_dgemm', description='A test program')
    call cli%add('-m' , 'Number of rows in the matrix.'                              , def=5             )
    call cli%add('-n' , 'Number of cols in the matrix. Equal to m if not specified.' , def=5             )
    call cli%add('-k' , 'Number of cols in the matrix. Equal to m if not specified.' , def=5             )
    call cli%add('-mb', 'Blocks size.'                                            , def=-1  )
    call cli%add('-ta'  , 'If set, factorize the transpose if m>=n.'                   , def='n' )
    call cli%add('-tb'  , 'If set, factorize the transpose if m>=n.'                   , def='n' )

    call cli%parse(info)
    if(info.ne.0) stop
    
    call cli%get('-m',  m)
    call cli%get('-n',  n)
    call cli%get('-k',  k)
    call cli%get('-mb', mb)
    call cli%get('-ta'  , tra)
    call cli%get('-tb'  , trb)

    transa = tra(1:1)
    transb = trb(1:1)
    
    return
    
  end subroutine get_cla
  
end program cqrm_dsgemm
