!! ##############################################################################################
!!
!! Copyright 2012-2024 CNRS, INPT
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################

program dqrm_dspotr

  use dqrm_mod
  implicit none

  type(dqrm_dsmat_type), target  :: a, rr, b, x, r
  type(dqrm_ws_type)             :: work
  type(qrm_dscr_type)            :: qrm_dscr

  integer                        :: info, i, j, k, it, nth 
  integer                        :: m, n, mb, ts, te, tr, verb
  real(kind(1.d0))               :: tfct, tslv, gflops
  logical                        :: pin=.true.
  real(r64)                      :: rnrm, bnrm, xnrm, err, anrm
  integer                        :: ncpu, ngpu
  
  call qrm_init()

  call get_cla(m, mb, verb)

  call qrm_get('qrm_ncpu', ncpu)
  call qrm_get('qrm_ngpu', ngpu)

  call qrm_log_record("Setup", verb=2)
  call qrm_log_record("- M","m","i10",m)
  call qrm_log_record("- MB","mb","i4",mb)
  call qrm_log_record("- NCPU","ncpu","i4",ncpu)
  call qrm_log_record("- NGPU","ngpu","i4",ngpu)
  if(verb.ge.2) call qrm_log_print('list',verb=verb)

  gflops = (real(m)**3)/3.d0/1e9

  call qrm_dsmat_init(a , m, m, mb=mb, info=info)
  call qrm_dsmat_init(rr, m, m, mb=mb, pin=pin, info=info)
  call qrm_dsmat_copy(a, rr, info=info)
  call qrm_ws_init(work, mb, mb)
  
  ! make a hpd matrix
  call qrm_dsmat_fill(rr, 'r')
  ! call qrm_dsmat_print(rr, 'f10.7', l=0)
  call qrm_dsmat_gemm(qrm_transp, qrm_no_transp, &
       qrm_done,                                         &
       rr, rr,                                           &
       qrm_dzero,                                        &
       a)

  call qrm_dsmat_copy(a, rr, info=info)

  call qrm_dsmat_nrm(a, anrm, info=info)

  ! call qrm_dsmat_print(rr, 'f15.10',ir=.true.)
  
  ! ########       ###        ######     ########     #######  
  ! ##            ## ##      ##    ##       ##       ##     ## 
  ! ##           ##   ##     ##             ##       ##     ## 
  ! ######      ##     ##    ##             ##       ##     ## 
  ! ##          #########    ##             ##       ##     ## 
  ! ##          ##     ##    ##    ##       ##       ##     ## 
  ! ##          ##     ##     ######        ##        #######  
  
  ! call qrm_dsmat_print(rr, 'f10.7', l=rr%m)
  call qrm_log_record("Factorization", verb=2)
  
  call qrm_dscr_init(qrm_dscr)
  call qrm_barrier()

  call qrm_start_profiling()
  call system_clock(ts)
  call qrm_dsmat_potrf_async(qrm_dscr, 'u', rr, work)
  call qrm_barrier()
  call system_clock(te,tr)
  call qrm_stop_profiling()
  
  ! call qrm_dsmat_print(rr, 'f10.7', l=rr%m)
  call qrm_dscr_destroy(qrm_dscr)
  tfct = max(real(te-ts)/real(tr),1e-20) ! the max is to avoid signaling divide by zero
  call qrm_log_record("- Time for factorization","Tf","es10.3",tfct)
  call qrm_log_record("- Gflops/s","gflops","f8.2",gflops/tfct)
  if(verb.ge.2) call qrm_log_print('list',verb=verb)

  !  ######      #######     ##          ##     ##    ######## 
  ! ##    ##    ##     ##    ##          ##     ##    ##       
  ! ##          ##     ##    ##          ##     ##    ##       
  !  ######     ##     ##    ##          ##     ##    ######   
  !       ##    ##     ##    ##           ##   ##     ##       
  ! ##    ##    ##     ##    ##            ## ##      ##       
  !  ######      #######     ########       ###       ########   

  call qrm_dsmat_init(b , m, 1, mb, info=info)
  call qrm_dsmat_init(r , m, 1, mb, info=info)
  call qrm_dsmat_init(x , m, 1, mb, info=info)
  call qrm_dsmat_fill(b, init='r')
  call qrm_dsmat_fill(r, init='z')
  call qrm_dsmat_copy(b, x, info=info)
  call qrm_dsmat_nrm(b, bnrm, info=info)

  call qrm_log_record("Solving Ax=b",verb=2)
  
  call system_clock(ts)
  call qrm_dsmat_trsm(qrm_left, qrm_upper, qrm_transp, qrm_no_unit, qrm_done, rr, x, info=info)
  call qrm_dsmat_trsm(qrm_left, qrm_upper, qrm_no_transp, qrm_no_unit, qrm_done, rr, x, info=info)
  call system_clock(te,tr)

  tslv = max(real(te-ts)/real(tr),1e-20) ! the max is to avoid signaling divide by zero
  
  call qrm_dsmat_nrm(x, xnrm, info=info)
  call qrm_dsmat_copy(b, r, info=info)
  call qrm_dsmat_gemm(qrm_no_transp, qrm_no_transp, qrm_done, a, x, qrm_dmone, r, info=info)
  call qrm_dsmat_nrm(r, rnrm, info=info)

  call qrm_log_record("- ||x||", "nrmx", "es12.5", xnrm, verb=1) 
  call qrm_log_record("- ||r||/||A||", "nrmr", "es12.5", rnrm/(anrm*xnrm+bnrm), verb=1) 
  call qrm_log_record("- Time for solve","time","es10.3",tslv)
  call qrm_log_record("=", verb=2)

  if(verb.ge.2) call qrm_log_print('list',verb=verb)
  if(verb.eq.1) call qrm_log_print('tab',verb=verb)

10 continue 

  ! cleanup
  
  call qrm_dsmat_destroy(a)
  call qrm_dsmat_destroy(rr)
  call qrm_dsmat_destroy(b)
  call qrm_dsmat_destroy(x)
  call qrm_dsmat_destroy(r)
  call qrm_ws_destroy(work)
     
  call qrm_finalize()
     
  stop


contains

  subroutine get_cla(m, mb, verb)
    use qrm_cli_mod
    implicit none

    integer            :: m, mb, verb, info
    type(qrm_cli_type) :: cli
    
    call cli%init(title='dqrm_dspotr', description='A test program for the Cholesky factorization of dense matrices')
    call cli%add('-m'  , 'Number of rows in the matrix.'                              , def=1024            )
    call cli%add('-mb' , 'Blocks rowsize.'                                            , def=qrm_default_mb_ )
    call cli%add('-v'  , 'Verbosity: 3=full, 2=batch w/ header, 1=batch w/o header.'  , def=3 )

    call cli%parse(info)

    call cli%get('-m'  , m)
    call cli%get('-mb' , mb)
    call cli%get('-v'  , verb)

    return
    
  end subroutine get_cla
  
end program dqrm_dspotr
