!! ##############################################################################################
!!
!! Copyright 2012-2024 CNRS, INPT
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!!##############################################################################################
!> sqrm_spata.F90
!!
!! An example program which reads a matrix and solves the
!! least-squares problem through QR factorization and then through
!! normal-equations+Cholesky
!!
!! ./sqrm_spata -mat matrix 
!!
!!
!!##############################################################################################

#include "qrm_common.h"

program sqrm_spata

  use sqrm_mod
  implicit none

  type(sqrm_spmat_type)          :: a_coo, a_csc, ata_csr, ata_coo, a_reg
  type(sqrm_spfct_type)          :: qrm_spfct

  integer                        :: info, i, j, ncpu, ngpu, ounit, verb
  integer(i64)                   :: ts, te, tr, nnzr, nnzh, gflopsqr, gflopsne
  real(r32)                      :: anrm, rnrm, bnrm, onrm, xnrm
  real(r32)                      :: ttqr, taqr, tfqr, tsqr, ttne, tane, tfne, tsne
  character(len=:), allocatable  :: matname
  real(r32), allocatable      :: b(:), x(:), r(:), atb(:), b_reg(:)
  real(r32)                      :: gamma
  
  call qrm_init()
  call qrm_stop_profiling()

  ! initialize the matrix data structure. 
  call qrm_spmat_init(a_coo)
  call qrm_stop_profiling()

  call qrm_get('qrm_ncpu', ncpu)
  call qrm_get('qrm_ngpu', ngpu)
  
  ! read the input file
  call qrm_read_parms(matname, a_coo, gamma, verb, info)
  __QRM_INFO_CHECK(info,'qrm_test','qrm_read_parm',20) 

  if(a_coo%m.lt.a_coo%n) call qrm_spmat_transpose(a_coo)
  call qrm_spmat_nrm(a_coo, 'f', anrm)

  call qrm_alloc(b   , a_coo%m)
  call qrm_alloc(r   , a_coo%m)
  call qrm_alloc(x   , a_coo%n)
  call qrm_alloc(atb , a_coo%n)

  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  call qrm_vec_larnv(b)
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! call qrm_vec_larnv(x)
  ! call qrm_spmat_mv(a_coo, qrm_no_transp, qrm_sone, x, qrm_szero, b)
  ! x = qrm_szero
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  call qrm_vecnrm(b, size(b,1), '2', bnrm)

  call qrm_spmat_copy(a_coo, a_reg)
  if(gamma.ne.0.0) then
     call qrm_tikhonov(a_reg, real(gamma*anrm,kind(a_reg%val(1))))
  end if


  call qrm_alloc(b_reg, a_reg%m)
  b_reg(1:a_coo%m)  = b
  b_reg(a_coo%m+1:) = qrm_szero

  call qrm_log_record("Setup", verb=2)
  call qrm_log_record("- Matrix name","mat","a16",matname)
  call qrm_log_record("- M","m","i10",a_coo%m, verb=2)
  call qrm_log_record("- N","n","i10",a_coo%n, verb=2)
  call qrm_log_record("- NNZ","nnz","i15",a_coo%nz, verb=2)
  call qrm_log_record("- NCPU","ncpu","i4",ncpu)
  call qrm_log_record("- NGPU","ngpu","i4",ngpu)
  if(verb.eq.2) call qrm_log_print('list',verb=verb)
  
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  !!
  !!   QR based least-squares
  !!
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  if(.true.) then
     call qrm_log_record("QR of A", verb=2)
     call qrm_log_record("- Type","tp","a2","qr")
     ttqr = 0.0
     call qrm_log_record("- Time for n-e","Tt","es10.3",ttqr)

     rnrm = -1.0
     onrm = -1.0
     
     call qrm_spfct_init(qrm_spfct, a_reg)
     call system_clock(ts)
     call qrm_analyse(a_reg, qrm_spfct, info=info)
     call system_clock(te, tr); taqr = real(te-ts)/real(tr)
     if(info.ne.0) goto 10
     call qrm_log_record("- Time for analysis","Ta","es10.3",taqr)
     
     call system_clock(ts)
     call qrm_factorize(a_reg, qrm_spfct, info=info)
     call system_clock(te, tr); tfqr = real(te-ts)/real(tr)
     if(info.ne.0) goto 10
     
     nnzr     = qrm_spfct%gstats(qrm_nnz_r_)
     nnzh     = qrm_spfct%gstats(qrm_nnz_h_)
     gflopsqr = qrm_spfct%gstats(qrm_e_facto_flops_)
     call qrm_log_record("- Time for factorization","Tf","es10.3",tfqr)
     call qrm_log_record("- Gflops at facto","f-gflops","f8.2",real(gflopsqr)/1e9)
     call qrm_log_record("- Gflops/s at facto","f-gflops/s","f8.2",real(gflopsqr)/tfqr/1e9)
     
     x = qrm_szero
     
     call system_clock(ts)
     call qrm_spfct_geqrs(qrm_spfct, b_reg, x, info=info)
     call system_clock(te, tr); tsqr = real(te-ts)/real(tr)
     if(info.ne.0) goto 10
     call qrm_log_record("- Time for solve","Ts","es10.3",tsqr)
     
     r = b
     
     call qrm_residual_norm(a_coo, r, x, rnrm)
     call qrm_vecnrm(x, size(x,1), '2', xnrm)
     call qrm_residual_orth(a_coo, r, onrm)   
     
     call qrm_spfct_destroy(qrm_spfct)

10 continue
     
     ! if(ounit.gt.0) write(ounit,'("| ",a16," | qr |",i5," |",i5," |",f8.2," |",f8.2," |",'// &
          ! 'f8.2," |",f8.2," |",f8.2," |",e12.5," |   ",e12.5," |")')&
          ! matname(1:min(len(matname),16))//repeat(' ',max(0,16-len(matname))), ncpu, ngpu, &
          ! real(gflopsqr)/1e9, ttqr, taqr, tfqr, tsqr, rnrm, onrm

     call qrm_log_record("- Residual norm","nrm-r","es10.3",rnrm)
     call qrm_log_record("- Max orth norm","nrm-o","es10.3",onrm)
     if(verb.eq.1) call qrm_log_print('tab',verb=verb, clean=.false.)
     if(verb.eq.2) call qrm_log_print('list',verb=verb)
     

  end if

  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  !!
  !!   Cholesky-based Normal Equations
  !!
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  call qrm_log_record("Cholesky of ATA", verb=2)
  call qrm_log_record("- Type","tp","a2","ch")
  rnrm = -1.d0
  onrm = -1.d0
  
  call system_clock(ts)
  call qrm_spmat_convert(a_reg, a_csc, 'csc')

  call qrm_spmat_ata(a_csc, ata_csr)
  call qrm_spmat_destroy(a_csc)
  call qrm_spmat_convert(ata_csr, ata_coo, 'coo')
  call qrm_spmat_destroy(ata_csr)
  call qrm_spmat_mv(a_coo, qrm_transp, qrm_sone, b, qrm_szero, atb)
  call system_clock(te, tr); ttne = real(te-ts)/real(tr)
  call qrm_log_record("- Time for n-e","Tt","es10.3",ttne)
  
  call qrm_spfct_init(qrm_spfct, ata_coo)
  call system_clock(ts)
  call qrm_analyse(ata_coo, qrm_spfct, info=info)
  call system_clock(te, tr); tane = real(te-ts)/real(tr)
  if(info.ne.0) goto 20
  call qrm_log_record("- Time for analysis","Ta","es10.3",tane)

  call system_clock(ts)
  call qrm_factorize(ata_coo, qrm_spfct, qrm_no_transp, info)
  call system_clock(te, tr); tfne = real(te-ts)/real(tr)
  if(info.ne.0) goto 20

  gflopsne = qrm_spfct%gstats(qrm_e_facto_flops_)
  call qrm_log_record("- Time for factorization","Tf","es10.3",tfne)
  call qrm_log_record("- Gflops at facto","f-gflops","f8.2",real(gflopsne)/1e9)
  call qrm_log_record("- Gflops/s at facto","f-gflops/s","f8.2",real(gflopsne)/tfne/1e9)

  call system_clock(ts)
  call qrm_spfct_potrs(qrm_spfct, atb, x, info=info)
  call system_clock(te, tr); tsne = real(te-ts)/real(tr)
  if(info.ne.0) goto 20
  call qrm_log_record("- Time for solve","Ts","es10.3",tsne)
  
  r = b

  call qrm_residual_norm(a_coo, r, x, rnrm)
  call qrm_vecnrm(x, size(x,1), '2', xnrm)
  call qrm_residual_orth(a_coo, r, onrm)   
  call qrm_log_record("- Residual norm","nrm-r","es10.3",rnrm)
  call qrm_log_record("- Max orth norm","nrm-o","es10.3",onrm)
  call qrm_log_record("=", verb=2)
  if(verb.eq.1)call qrm_log_print('tab',verb=1, hdr=.false.)
  if(verb.eq.2) call qrm_log_print('list',verb=verb)

  call qrm_spfct_destroy(qrm_spfct)

20 continue
  
  call qrm_spmat_destroy(ata_coo)
  call qrm_spmat_destroy(a_coo)
  call qrm_spmat_destroy(a_reg)
  
  call qrm_finalize()
  stop

contains

  subroutine qrm_read_parms(matname, qrm_spmat, gamma, verb, info)
    ! Function: read_parms
    !
    ! *Input*:
    !
    ! *Output*:
    !
    use qrm_ssget_mod
    use sqrm_utils_mod
    use qrm_cli_mod
    implicit none

    type(sqrm_spmat_type)         :: qrm_spmat
    character(len=:), allocatable :: matname
    real(r32)                     :: gamma
    integer                       :: verb
    integer, optional             :: info

    type(qrm_cli_type)            :: cli
    character(len=:), allocatable :: mat, matfile, ssrepo, ssmat
    integer                       :: col1, col2, col3, col4, err
    integer                       :: nx, ny, nz, n, s, ounit
    real(r32)                     :: anrm
    
    call cli%init(title='sqrm_spata', description='A test program for computing A^T*A')
    call cli%add('-mat'  , 'The matrix file/name.\n'//&
         'This can either be a file in MatrixMarket format, \n'//&
         'a string of the form ss:path:matname, in which case \nmatrix '//&
         'matname from SuiteSparse will be downloaded \nin path and used '//&
         'or a string of the type Mitt3D_27:N \nwhere a Mitt3D_27 matrix of '//&
         'size N will be generated.', def='', req=.true.)
    call cli%add('-t'    , 'Apply Tikhonov regularization.'                             , def=0.0_r32)
    call cli%add('-v'    , 'Verbosity: 3=full, 2=batch w/ header, 1=batch w/o header.'  , def=2 )

    call cli%parse(info)
    if(info.ne.0) then
       stop
    end if
    
    call cli%get('-mat'  , mat)
    call cli%get('-t'    , gamma)
    call cli%get('-v'    , verb)

    if(index(mat, 'ss:').eq.1) then
       ! this is a SuiteSparse matrix
       col1 = index(mat, ':')
       if(col1.eq.0) then
          write(*,'("SS matrix not correctly formatted.")')
          stop
       end if
       col2 = index(mat(col1+1:),':')+col1
       if(col2.eq.col1 .or. col2.eq.col1+1) then
          ssrepo = '.'
       else
          ssrepo = mat(col1+1:col2-1)
       end if
       call qrm_ssget_init(ssrepo)
       matfile = qrm_ssget(mat(col2+1:), err)
       __QRM_INFO_CHECK(err,'qrm_test','qrm_ssget',20)
       call qrm_readmat(matfile, qrm_spmat, .true., err)
       __QRM_INFO_CHECK(err,'qrm_test','qrm_readmat',20)
       matname = mat(col2+1:)
       ! else if(index(mat, '7PT:').eq.1) then
       ! this is a 7PT matrix
       ! col1 = index(mat, ':')
       ! col2 = index(mat(col1+1:),':')+col1
       ! col3 = index(mat(col2+1:),':')+col2
       ! col4 = index(mat(col3+1:),':')+col3
       ! if(col1.eq.0 .or. col2.eq.col1 .or. col3.eq.col2 .or. col4.eq.col3) then
          ! write(*,'("7PT matrix not correctly formatted.")')
          ! stop
       ! end if
       ! read(mat(col1+1:col2-1),*)nx
       ! read(mat(col2+1:col3-1),*)ny
       ! read(mat(col3+1:col4-1),*)nz
       ! read(mat(col4+1:      ),*)s
       ! call qrm_matgen_7pt(qrm_spmat, nx, ny, nz, s)
    else if(index(mat, 'Mitt3D_27:').eq.1) then
       ! this is a Mitt3D matrix
       col1 = index(mat, ':')
       if(col1.eq.0) then
          write(*,'("Mitt3D_27 matrix not correctly formatted.")')
          stop
       end if
       read(mat(col1+1:),*)n
       call qrm_matgen_mitt3d(qrm_spmat, n)
       matname = mat
    else
       matfile = mat
       call qrm_readmat(matfile, qrm_spmat, .true., err)
       __QRM_INFO_CHECK(err,'qrm_test','qrm_readmat',20) 
       matname = mat
    end if

20  continue
    if(present(info)) info = err
    return
    
  end subroutine qrm_read_parms


  
end program sqrm_spata






